from input_for_voyage_model import *
import itertools
from models.route_model import route_model
from models.route_model_with_y import route_model_with_y
from models.route_model_columns import route_model_columns
from models.fix_y_model import fix_y_model
from copy import deepcopy
from random import shuffle, randint, choice, random
from config import BUFFER, PRINT_STUFF, BUFFER1
from math import ceil


def get_visit_load_unload_port(routes, route_visits_port, contracts, contract_ports):
    visit_load_unload_port = dict()
    for c in contracts:
        for r in routes:
            if route_visits_port[contract_ports[c][0], r] and route_visits_port[contract_ports[c][1], r]:
                visit_load_unload_port[c, r] = 1
            else:
                visit_load_unload_port[c, r] = 0
    return visit_load_unload_port


def get_pickups(contracts, minimum_pickups):
    pickups = dict()
    for c in contracts:
        if minimum_pickups[c] > 0:
            pickups[c] = minimum_pickups[c]
        else:
            pickups[c] = 1
    return pickups


def shuffle_routes_to_voyage(sol_routes):
    return shuffle(sol_routes[0])


def get_y_voyage_vessel(y_model, sol_routes):
    y = dict()
    route_and_vessel = []
    for var in y_model.variables.get_names():
        if y_model.solution.get_values(var) > 0.9:
            sol = get_solution_var(var, "Y")
            route_and_vessel.append((int(sol[1]), sol_routes[int(sol[0])]))
            y[int(sol[0]) + 1, int(sol[1])] = 1
    return y, route_and_vessel


def get_solution_routes(sol_routes):
    solution_routes = dict()
    for v in voyages:
        if v > len(sol_routes[0]):
            solution_routes[v] = 0
        else:
            solution_routes[v] = sol_routes[0][v - 1]
    return solution_routes


def get_solutions(MODEL, problem):
    sol_routes = get_route_solutions(problem, MODEL)
    sol_vars = []
    sol_vals = []
    for var in problem.variables.get_names():
        if problem.solution.get_values(var) > 0.1:
            sol_vars.append(var)
            sol_vals.append(int(problem.solution.get_values(var)))

    solution_routes = get_solution_routes(sol_routes)

    for x in range(len(voyages)-len(sol_routes[0])):
        sol_routes[0].append(0)
    return sol_routes, solution_routes, sol_vars, sol_vals


def print_routes(sol_routes, routes):
    for el in sol_routes[0]:
        print(routes[el])


def get_quantity_per_port(demand_in_port):
    demand_per_port = dict()
    for i in ports:
        demand_per_port[i] = 0
    for key, value in demand_in_port.items():
        demand_per_port[key[0]] += value
    return demand_per_port


def get_average_capacity():
    return sum(vessel_capacity.values())/(len(vessels))


def get_average_capacity_product():
    average_capacity_products = dict()
    for p in products:
        average_capacity_products[p] = 0
    for key, value in vessel_capacity.items():
        average_capacity_products[key[1]] += value
    for key, value in average_capacity_products.items():
        average_capacity_products[key] = value/vessel_count
    return average_capacity_products


def get_demand_in_port():
    demand_in_port = dict()
    for i in ports:
        for p in products:
            dem = 0
            for c in contracts_L[i]:
                if (c, p) in contract_demand:
                    dem += contract_demand[c, p]
            demand_in_port[i, p] = dem
    return demand_in_port


def create_routes():
    lst = [list(i) for i in itertools.product([0, 1], repeat=len(ports))]
    routes = dict()
    i = 0
    for el in lst:
        if sum(el) != 1:
            if check_load_to_unload(el):
                routes[i] = el
                i += 1
    return routes


def check_load_to_unload(route):
    load = 0
    unload = 0
    for i, el in enumerate(route):
        index = i + vessel_count + 1
        if el == 1:
            if contracts_L[index]:
                load = 1
            if contracts_U[index]:
                unload = 1
    if load+unload == 1:
        return False
    return True


def find_visits(route):
    visit = []
    for i, el in enumerate(route):
        if el == 1:
            visit.append(i+vessel_count+1)
    return visit


def visits_for_routes(routes):
    visits = dict()
    for key, value in routes.items():
        visits[key] = find_visits(value)
    return visits


def time_routes(sol_routes, visits, distance_in):
    time_route = dict()
    index = 1
    for route in sol_routes[0]:
        time_route.setdefault(index, dict())
        port = visits[route]
        for i in range(len(port)-1):
            cost = 0
            for j in range(i+1, len(visits[route])):
                cost += distance_in[port[j-1], port[j]]
                time_route[index][port[i], port[j]] = cost
        index += 1
    return time_route


def get_sailing_time(load_port, unload_port, y, voyage, time_route, product, load, c):
    return time_route[voyage][load_port, unload_port] * sailing_time[find_vessel(voyage, y), velocities_count] \
           + node_time[load_port] + handling_times[product] * load


def calculate_cost(key, visits):
    cost = 0
    port = visits[key]
    for i in range(len(visits[key])-1):
        cost += distance_in[port[i], port[i+1]]
        cost += port_cost[port[i]]
    if len(visits[key]) > 1:
        cost += port_cost[port[len(port) - 1]]
    return cost


def cost_of_route(routes, visits):
    cost_of_route = dict()
    for key, value in routes.items():
        cost_of_route[key] = calculate_cost(key, visits)
    return cost_of_route


def find_minimum_visit():
    minimum_visits_port = dict()
    minimum_visits_products = dict()
    for p in ports:
        min_visit = 0
        for c in contracts_L[p]:
            if min_visit < minimum_pickups[c]:
                min_visit = minimum_pickups[c]
            for prod in products:
                if (c, prod) in contract_demand:
                    if (p, prod) in minimum_visits_products and minimum_pickups[c] > \
                       minimum_visits_products[p, prod]:
                        minimum_visits_products[p, prod] = minimum_pickups[c]
                    if (p, prod) not in minimum_visits_products:
                        minimum_visits_products[p, prod] = minimum_pickups[c]
        for c in contracts_U[p]:
            if min_visit < minimum_pickups[c]:
                min_visit = minimum_pickups[c]
            for prod in products:
                if (c, prod) in contract_demand:
                    if (p, prod) in minimum_visits_products and minimum_pickups[c] > \
                       minimum_visits_products[p, prod]:
                        minimum_visits_products[p, prod] = minimum_pickups[c]
                    if (p, prod) not in minimum_visits_products:
                        minimum_visits_products[p, prod] = minimum_pickups[c]
        minimum_visits_port[p] = min_visit
    return minimum_visits_port, minimum_visits_products


def if_route_visits_port(routes, visits):
    route_visits_port = dict()
    for r in routes:
        for i in ports:
            if i in visits[r]:
                route_visits_port[i, r] = 1
            else:
                route_visits_port[i, r] = 0
    return route_visits_port


def get_solution_var(variable, var):
    return variable.strip(var).split(",")


def get_routes_from_model(MODEL, cost_of_route, routes, minimum_visits_port, route_visits_port, ports,
                          voyage_count, vessels, demand_in_port, average_capacity, demand_per_port,
                          average_capacity_product, visit_load_unload_port, pickups, min_sail_time_contract_route):
    if MODEL == 1:
        problem = route_model(cost_of_route, routes, minimum_visits_port, route_visits_port, ports,
                              voyage_count, vessels, demand_in_port, average_capacity, demand_per_port,
                              average_capacity_product, visit_load_unload_port, pickups, min_sail_time_contract_route)
    elif MODEL == 2:
        problem = route_model_with_y(cost_of_route, routes, minimum_visits_port, route_visits_port, ports,
                                     voyage_count, vessels, demand_in_port)
    else:
        problem = route_model_columns(routes, minimum_visits_port, route_visits_port, ports, voyage_count,
                                      vessels, demand_in_port)
    return problem


def get_route_solutions(problem, MODEL):
    solution_routes = []
    y = dict()
    voyage = 1

    for var in problem.variables.get_names():
        if problem.solution.get_values(var) > 0:
            for x in range(round(problem.solution.get_values(var))):
                sol = get_solution_var(var, "X")
                solution_routes.append(int(sol[0]))
            if MODEL == 2:
                y[voyage, int(sol[1])] = 1
                voyage += 1
    return solution_routes, y


def get_y(route_and_vessel, routes):
    y = dict()
    v = 1
    for vessel, route in route_and_vessel:
        if route in route_and_vessel.values():
            y[routes[route], vessel] = 1
        v += 1
    return y


def get_demands_and_min_pickup(visit_load_unload_port, sol_routes, min_sail_time_contract_route):
    demands = dict()
    min_pickup = dict()

    for p in products:
        demands.setdefault(p, dict())
        min_pickup.setdefault(p, dict())
        for c in contracts:
            visits = 0
            for route in sol_routes:
                cont = True
                if c in transit_contracts and (c, route) in min_sail_time_contract_route and min_sail_time_contract_route[c, route] != 1:
                    cont = False
                if visit_load_unload_port[c, route] and cont:
                    visits += 1
            if (c, p) in contract_demand:
                if minimum_pickups[c] > 1:
                    load = minimum_quantity[c, p]
                else:
                    load = contract_demand[c, p]
                demands[p][c] = contract_demand[c, p]
                pickup = minimum_pickups[c] if minimum_pickups[c] > 0 else 1
                transit = -100
                options = pickup - visits
                min_pickup[p][c] = (options, transit, load)
    return demands, min_pickup

def check_transit_time_route(contract, voyage, load, time_routes):
    cont = True
    # Check transit time requirement
    product = get_product_for_contract(contract)
    sail_time = time_routes[voyage][load_port, unload_port] * sailing_time[1, velocities_count] \
           + node_time[load_port] + handling_times[product] * load
    if sail_time > contract_transit_time[contract]:
        cont = False
    return cont


def get_product_capacity():
    product_capacity = dict()
    for p in products:
        product_capacity.setdefault(p, dict())
    for key, value in vessel_capacity.items():
        product_capacity[key[1]][key[0]] = value
    return product_capacity


def get_route_and_vessel(sol_routes, MODEL, cost_of_route):
    y = dict()
    if MODEL != 2:
        y_model = fix_y_model(sol_routes[0], cost_of_route)
        y_model.solve()
        for var in y_model.variables.get_names():
            if y_model.solution.get_values(var):
                sol = get_solution_var(var, "Y")
                y[int(sol[0])] = int(sol[1])
    else:
        y = sol_routes[1]
    return y


# Returns the vessel assigned to the voyage
def find_vessel(voyage, y):
    for k in vessels:
        if (voyage, k) in y:
            return k


def check_transit_time(contract, voyage, load, time_routes, y):
    cont = True
    # Check transit time requirement
    product = get_product_for_contract(contract)
    sail_time = get_sailing_time(load_port[contract], unload_port[contract], y, voyage, time_routes, product,
                                 load, contract)
    if sail_time > contract_transit_time[contract]:
        cont = False
    return cont


def fix_delta(possible_route, voy_contract, capacity, space_used, c, y, contract_vessel, voyage_used, load, time_routes):
    for voyage1 in possible_route[c]:
        for contract in voy_contract[voyage1]:
            for voyage in possible_route[contract[0]]:
                if contract not in voy_contract[voyage]:
                    cont = True
                    prod = get_product_for_contract(contract[0])
                    if contract[0] in transit_contracts:
                        cont = check_transit_time(contract[0], voyage, load, time_routes, y)

                    if contract[1] < vessel_capacity[find_vessel(voyage, y), prod] - space_used[find_vessel(voyage, y)] \
                            and ((voyage, contract[0]) not in contract_vessel) and cont: # and voyage1 in voyage_used:
                        #print("FIX DELTA: ", contract[0], "to", voyage, "from", voyage1)
                        contract_vessel[voyage1, contract[0]] = 0
                        contract_vessel[voyage, contract[0]] = 1
                        voyage_used.remove(voyage1)
                        #voyage_used.append(voyage)
                        voy_contract[voyage1].remove(contract)
                        voy_contract[voyage].append(contract)
                        space_used[find_vessel(voyage1, y)] -= contract[1]
                        # NEW
                        space_used[find_vessel(voyage, y)] += contract[1]
                        possible_route[contract[0]].append(voyage1)
                        return contract_vessel, voyage_used, voyage1, space_used, possible_route
    return False, False, False, False, possible_route


# Returns all routes that visit both the loading and unloading port of the contract
def get_possible_routes_for_contract(c, sol_rou, visit_load_unload_port):
    possible_route = []
    for i, route in enumerate(sol_rou):
        if visit_load_unload_port[c, route]:
            possible_route.append((i + 1))
    return possible_route


# TODO: MAKE DELTA GREAT (AGAIN)
def place_contract_help(load, capacity, voyage_used, c, contract_vessel, space_used, voyage_to_be_taken,
                        route_visits_port, sol_routes, y, voy_available, time_routes, product, visit_load_unload_port, sol_rou, possible_route, voy_contract, transit_voyages, not_transit_voyages):
    unloaded = True
    full = []
    while unloaded:
        voyage = 0
        most_space = 0
        min_speed = 10000000000000
        for i, route in enumerate(sol_rou):
            if (i+1) in possible_route[c]:
                if PRINT_STUFF:
                    print("Voyage ", i+1, capacity[find_vessel(i + 1, y)], space_used[find_vessel(i + 1, y)])
                available_space = capacity[find_vessel(i + 1, y)] - space_used[find_vessel(i + 1, y)]
                if available_space > most_space and (i + 1) not in voyage_used and c not in transit_contracts:
                    most_space = available_space
                    voyage = i + 1
                if c in transit_contracts and (i + 1) not in voyage_used:
                    sail_time = get_sailing_time(load_port[c], unload_port[c], y, i+1, time_routes, product, load, c)
                    if sail_time < min_speed:
                        voyage = i + 1
                        min_speed = sail_time


        if len(voyage_used) == voyage_count or voyage == 0:
            if len(possible_route[c]) > 0:
                contract_vessel, voyage_used, voyage, space_used, possible_route = fix_delta(possible_route,
                                                                                             voy_contract, capacity,
                                                                                             space_used, c, y,
                                                                                             contract_vessel,
                                                                                             voyage_used, load,time_routes)
                if contract_vessel == False:
                    return voyage_used, [], space_used, voyage_to_be_taken, voy_available, voy_contract, transit_voyages, not_transit_voyages
            else:
                return voyage_used, [], space_used, voyage_to_be_taken, voy_available, voy_contract, transit_voyages, not_transit_voyages


        cont = True
        # Check transit time requirement
        if c in transit_contracts and route_visits_port[load_port[c], sol_routes[voyage]] == 1 and \
                route_visits_port[unload_port[c], sol_routes[voyage]] == 1:
            sail_time = get_sailing_time(load_port[c], unload_port[c], y, voyage, time_routes, product, load, c)
            if sail_time > contract_transit_time[c]:
                cont = False

        # Check capacity requirement
        if voyage != 0 and load <= (capacity[find_vessel(voyage, y)] - space_used[find_vessel(voyage, y)]) and \
           voyage not in voyage_used and cont:
            space_used[find_vessel(voyage, y)] += load
            contract_vessel[voyage, c] = 1
            unloaded = False
            possible_route[c].remove(voyage)
            voy_contract[voyage].append((c, load))
            if voyage not in voyage_to_be_taken:
                voyage_to_be_taken.append(voyage)
        elif voyage not in full:
            full.append(voyage)
        if voyage not in voyage_used:
            voyage_used.append(voyage)
    return voyage_used, contract_vessel, space_used, voyage_to_be_taken, voy_available, voy_contract, transit_voyages, \
           not_transit_voyages


def place_contracts(product, demand_product, capacity, min_pickups, space_used, voyage_to_be_taken, route_visits_port,
                    sol_routes, y, time_routes, visit_load_unload_port, sol_rou, possible_route, voy_contract,
                    transit_voyages, not_transit_voyages):
    delta = dict()
    contract_vessel = dict()
    for c in sorted(min_pickups.keys(), key=lambda x: min_pickups[x], reverse=True):
        #print("Contract", c, minimum_pickups[c] )
        possible_route[c] = get_possible_routes_for_contract(c, sol_rou, visit_load_unload_port)
        voy_available = deepcopy([k for k in sorted(sol_routes, key=sol_routes.get)])
        voyage_used = []
        if minimum_pickups[c] > 1:
            load = minimum_quantity[c, product]
            if load <= maximum_quantity[c, product]:
                spread = minimum_pickups[c]
                for t in range(spread):
                    voyage_used, contract_vessel, space_used, voyage_to_be_taken, voy_available, voy_contract, \
                    transit_contracts, not_transit_voyages = place_contract_help(load, capacity, voyage_used, c,
                                                                                 contract_vessel, space_used,
                                                                                 voyage_to_be_taken, route_visits_port,
                                                                                 sol_routes, y, voy_available,
                                                                                 time_routes, product,
                                                                                 visit_load_unload_port, sol_rou,
                                                                                 possible_route, voy_contract,
                                                                                 transit_voyages, not_transit_voyages)
                    if contract_vessel == []:
                        return False, delta, space_used, voyage_to_be_taken, possible_route, transit_voyages, \
                               not_transit_voyages
        else:
            load = contract_demand[c, product]
            if load <= maximum_quantity[c, product]:
                voyage_used, contract_vessel, space_used, voyage_to_be_taken, voy_available, voy_contract, \
                transit_voyages, not_transit_voyages = place_contract_help(load, capacity, voyage_used, c,
                                                                           contract_vessel, space_used,
                                                                           voyage_to_be_taken, route_visits_port,
                                                                           sol_routes, y, voy_available, time_routes,
                                                                           product, visit_load_unload_port, sol_rou,
                                                                           possible_route, voy_contract,
                                                                           transit_voyages, not_transit_voyages)
                if contract_vessel == []:
                    return False, delta, space_used, voyage_to_be_taken, possible_route, transit_voyages, \
                           not_transit_voyages
    delta = contract_vessel
    return True, delta, space_used, voyage_to_be_taken, possible_route, transit_voyages, not_transit_voyages


def get_delta(demands, product_capacity, min_pickup, route_visits_port, solution_routes, y, time_route, visit_load_unload_port, sol_rou):
    d1 = dict()
    voyage_to_be_taken = []
    space_used = dict()
    possible_route = dict()
    voy_contract = dict()
    transit_voyages = []
    not_transit_voyages = [x for x in range(1, len(voyages)+1)]
    for v in voyages:
        voy_contract.setdefault(v, [])

    for k in vessels:
        space_used[k] = 0
    for p in reversed(products):
        check, delta, space_used, voyage_to_be_taken, possible_route, transit_voyages, not_transit_voyages = \
            place_contracts(p, demands[p], product_capacity[p], min_pickup[p], space_used, voyage_to_be_taken,
                            route_visits_port, solution_routes, y, time_route, visit_load_unload_port, sol_rou,
                            possible_route, voy_contract, transit_voyages, not_transit_voyages)
        if check:
            d1.update(delta)
        else:
            return {}, space_used
    return d1, space_used


# Preprocessing: Fixing the variables (x, y, delta, z and phi)
def run_model(delta, y, contracts):
    voy = {}
    for v in voyages:
        voy.setdefault(v, [])
        for c in contracts:
            if (v, c) in delta and delta[v, c] == 1 and load_port[c] not in voy[v]:
                voy[v].append(load_port[c])
            if (v, c) in delta and delta[v, c] == 1 and unload_port[c] not in voy[v]:
                voy[v].append(unload_port[c])

    x_var = {}
    z_var = {}
    for v in voyages:
        ports = sorted(voy[v])
        for k in vessels:
            if (v, k) in y and y[v, k] == 1:
                if len(ports) > 0:
                    x_var[k, ports[0], v] = 1
                    x_var[ports[len(ports)-1], last_port+k, v] = 1
                else:
                    x_var[k, last_port+k, v] = 1
        for i in range(len(ports)-1):
            x_var[ports[i], ports[i+1], v] = 1

    c_var = {}
    for c in evenly_spread_contracts:
        c_var.setdefault(c, [])
        for v in voyages:
            if (v, c) in delta and delta[v, c] == 1:
                c_var[c].append(v)

    for c in evenly_spread_contracts:
        contracts = sorted(c_var[c])
        for i in range(len(c_var[c])-1):
            z_var[contracts[i], contracts[i+1], c] = 1

    phi = {}
    n_var = {}
    for c in evenly_spread_contracts:
        n_var[c] = 0
        for v in voyages:
            if (v, c) in delta and delta[v, c] == 1:
                n_var[c] += 1

    for c, n in n_var.items():
        phi[n, c] = 1
    return x_var, delta, phi, z_var, y


def print_solution_voyage_model(problem2, best_solution):
    print("----------------------------------")
    print("Status: ", problem2.solution.get_status_string(problem2.solution.get_status()))
    print("Objective: ", "{:,}".format(round(problem2.solution.get_objective_value())))
    print("Best solution: ", "{:,}".format(best_solution))
    print("----------------------------------")


def write_solution_to_file(FILENAME, problem2):
    with open('results/full/' + FILENAME, 'a') as f:
        for var in problem2.variables.get_names():
            if problem2.solution.get_values(var) != 0:
                f.write(str(var) + "\t")
                f.write(str(problem2.solution.get_values(var)) + "\n")
        f.write("\n")


def get_distances_between_ports_for_route(visits_for_routes, routes):
    distances_between_ports_in_route = dict()
    for r, value in routes.items():
        distances_between_ports_in_route.setdefault(r, dict())
        index = 0
        port = visits_for_routes[r]
        for i in range(len(port) - 1):
            cost = 0
            for j in range(i + 1, len(visits_for_routes[r])):
                cost += distance_in[port[j-1], port[j]]
                distances_between_ports_in_route[r][port[i], port[j]] = cost
        index += 1
    return distances_between_ports_in_route


def get_min_sail_time_contract_route(visits_for_routes, routes):
    distances_between_ports_in_route = get_distances_between_ports_for_route(visits_for_routes, routes)
    min_sail_time_contract_route = dict()
    for c in transit_contracts:
        for r in routes:
            if load_port[c] in visits_for_routes[r] and unload_port[c] in visits_for_routes[r]:
                count_port = 0
                for p in range(load_port[c], unload_port[c] + 1):
                    if p in visits_for_routes[r]:
                        count_port += 1
                prod = get_product_for_contract(c)
                if minimum_pickups[c] > 1:
                    load = minimum_quantity[c, prod]
                else:
                    load = contract_demand[c, prod]
                sail_time = distances_between_ports_in_route[r][load_port[c], unload_port[c]] * \
                            sailing_time[1, velocities_count] + node_time[load_port[c]] * count_port + \
                            handling_times[prod] * load
                if contract_transit_time[c] > sail_time:
                    min_sail_time_contract_route[c, r] = 1
    return min_sail_time_contract_route


def search_delta(s_var, s_index, delta, sol_rou, visit_load_unload_port, available_space):
    switched = False
    s = int(s_index.split("S")[1])
    possible_routes = get_possible_routes_for_contract(s, sol_rou, visit_load_unload_port)
    taken_routes = []
    for v in voyages:
        if (v, s) in delta:
            taken_routes.append(v)
    if len(taken_routes) >= maximum_pickups[c]:
        return delta, switched
    diff_routes = list(set(possible_routes).symmetric_difference(taken_routes))
    change = random()
    while len(diff_routes) > 0:
        new_voy = choice(diff_routes)
        diff_routes.remove(new_voy)
        prod = get_product_for_contract(s)
        load = minimum_quantity[s, prod]
        if available_space[new_voy][prod] >= load:
            delta[new_voy, s] = 1
            switched = True
            taken_routes.append(new_voy)
            if change > 0.9 or len(taken_routes) > maximum_pickups[c]:
                delta[choice(taken_routes), s] = 0
    return delta, switched


# Find
def check_delta_transit_contracts(delta, y):
    contracts_time = []
    contracts_violating = []
    for c in transit_contracts:
        for v in voyages:
            check, time_diff, delta, v = check_sailing_time(delta, y, c, v)
            if v:
                contracts_time.append((c, v, time_diff))
            if not check:
                contracts_violating.append((c, v, time_diff))
    return contracts_violating, delta, contracts_time


# Returns dict with all sailing times for voyage
def check_sailing_time(delta, y, c, v):
    diff = 0
    loads = find_voyage_loads_per_port(v, delta)
    port_visits = get_port_visits(loads)
    distances_between_ports = get_ditances_between_ports(v, port_visits)
    if (v, c) in delta and delta[v, c] == 1:
        l_port = load_port[c]
        u_port = unload_port[c]
        time = get_sailing_time_contract(l_port, u_port, loads, port_visits, distances_between_ports, v, y)
        diff = time - contract_transit_time[c]
        if time >= contract_transit_time[c]:
            for voy in voyages:
                if voy != v:
                    loads = find_voyage_loads_per_port(voy, delta)
                    port_visits = get_port_visits(loads)
                    distances_between_ports = get_ditances_between_ports(voy, port_visits)
                    '''if (l_port, u_port) in distances_between_ports:
                        time_voy = get_sailing_time_contract(l_port, u_port, loads, port_visits,
                                                             distances_between_ports, voy, y)
                        if time_voy <= contract_transit_time[c] and ((voy, c) not in delta or delta[voy, c] == 0):
                            delta[v, c] = 0
                            delta[voy, c] = 1
                            print("CHECK SAILING TIME: Switch contract", c, "to voyage", voy)
                            return False, time_voy - contract_transit_time[c], delta, voy'''
            return False, diff, delta, v
        return True, diff, delta, v
    return True, diff, delta, None


def get_sailing_time_contract(l_port, u_port, loads, port_visits, distances_between_ports, voyage, y):
    time = distances_between_ports[l_port, u_port] * sailing_time[find_vessel(voyage, y), velocities_count]
    for i in range(l_port, u_port):
        if i in port_visits:
            time += node_time[i]
        for p in products:
            time += loads[i, p] * handling_times[p]
    return time


def get_ditances_between_ports(v, port):
    distances_between_ports_in_voyage = dict()
    for i in range(len(port) - 1):
        cost = 0
        for j in range(i + 1, len(port)):
            cost += distance_in[port[j - 1], port[j]]
            distances_between_ports_in_voyage[port[i], port[j]] = cost
    return distances_between_ports_in_voyage


def get_port_visits(loads):
    port_visits = []
    for i in ports:
        for p in products:
            if loads[i, p] > 0 and i not in port_visits:
                port_visits.append(i)
    return port_visits


def find_voyage_loads_per_port(voyage, delta):
    load_port_i = dict()
    contracts_on_voyage = get_contracts_on_voyage(voyage, delta)
    for i in ports:
        for p in products:
            load_port_i[i, p] = 0
        for c in contracts_on_voyage:
            if c in contracts_L[i] or c in contracts_U[i]:
                for p in products:
                    if (c, p) in contract_demand:
                        if minimum_pickups[c] > 1:
                            load_port_i[i, p] += contract_demand[c, p] / minimum_pickups[c]
                        else:
                            load_port_i[i, p] += contract_demand[c, p]
    return load_port_i


def get_contracts_on_voyage(voyage, delta):
    list = []
    for d in delta:
        if d[0] == voyage and delta[d] == 1:
            list.append(d[1])
    return list


def repair_transit_time_contract(violating_contracts, delta, sol_rou, visit_load_unload_port, space_used,
                                 product_capacity, y, available_space, ex_contracts):
    handling_times_for_contract = get_handling_time_for_contract()
    handling_times_for_contract = sorted(handling_times_for_contract, reverse=True)

    for (c, voy, time) in violating_contracts:
        l_port = load_port[c]
        u_port = unload_port[c]
        contracts_on_voy = get_contracts_on_voyage(voy, delta)
        for product in reversed(products):
            for time, new_c in handling_times_for_contract:
                if new_c in contracts_on_voy:
                    if ((load_port[new_c] >= l_port and load_port[new_c] < u_port) or
                            (unload_port[new_c] >= l_port and unload_port[new_c] < u_port)) \
                            and new_c not in ex_contracts:
                        possible_routes = get_possible_routes_for_contract(new_c, sol_rou, visit_load_unload_port)
                        prod = get_product_for_contract(new_c)
                        if minimum_pickups[c] <= 1:
                            load = contract_demand[new_c, prod]
                        else:
                            load = minimum_quantity[new_c, prod]
                        for route in reversed(possible_routes):
                            extra_spread = False
                            capacity = available_space[route][prod]
                            if ((route, new_c) not in delta or delta[route, new_c] == 0) \
                                    and new_c not in transit_contracts and capacity > load:
                                delta[route, new_c] = 1
                                if not extra_spread:
                                    delta[voy, new_c] = 0
                                    space_used[find_vessel(voy, y)] -= load
                                space_used[find_vessel(route, y)] += load
                                #print("REPAIR: ", new_c, "moved to voyage", route, time)
                                ex_contracts.append(new_c)
                                return delta, True, space_used, ex_contracts
    return delta, False, space_used, ex_contracts


def get_handling_time_for_contract():
    handling_time_for_contract = []
    for c in contracts:
        min_load = 0
        for p in products:
            if (c, p) in contract_demand:
                if minimum_pickups[c] > 1:
                    min_load = minimum_quantity[c, p]
                else:
                    min_load = contract_demand[c, p]
                handling_time_for_contract.append((min_load * handling_times[p], c))
    return handling_time_for_contract


# Add extra spread
'''def get_spread_max_min():
    max_min_spread = dict()
    for c in contracts:
        max_min_spread[c] = maximum_pickups[c] - minimum_pickups[c]
    return max_min_spread


def get_max_diff_spread(excluding):
    max_min_spread = get_spread_max_min()
    print("MAX_MIN_SPREAD: ", max_min_spread)
    max_diff_spread = 0
    contract = 0
    for c in contracts:
        if max_min_spread[c] > max_diff_spread and c not in excluding:
            max_diff_spread = max_min_spread[c]
            contract = c
    print("CONTRACT", contract)
    return contract'''


def add_extra_spread(c, delta,  sol_rou, visit_load_unload_port, available_space, product_capacity, y):
    possible_routes = get_possible_routes_for_contract(c, sol_rou, visit_load_unload_port)
    for p in products:
        if (c, p) in contract_demand:
            prod = p
    count = 0
    for v in voyages:
        if (v, c) in delta and delta[v, c] == 1:
            count += 1
    for v in possible_routes:
        if (v, c) not in delta or delta[v, c] == 0:
            if available_space[v][p] >= minimum_quantity[c, prod] and count < maximum_pickups[c]:
                print("EXTRA SPREAD", v, c)
                delta[v, c] = 1
                return delta, True
    print("NOT POSSIBLE TO ADD EXTRA SPREAD")
    return delta, False


def get_loads(delta, sol_rou, visit_load_unload_port, space_used, product_capacity, y):
    contracts_on_voyage = dict()
    total_load = dict()

    contracts_demand_spread = dict()
    for (c, p), demand in contract_demand.items():
        contracts_demand_spread[c] = demand

    for v in voyages:
        total_load[v] = 0
        contracts_on_voyage[v] = get_contracts_on_voyage(v, delta)
        for c in contracts_on_voyage[v]:
            for p in products:
                if (c, p) in contract_demand:
                    if minimum_pickups[c] <= 1:
                        total_load[v] += contract_demand[c, p]
                        contracts_demand_spread[c] -= contract_demand[c, p]
                    else:
                        total_load[v] += minimum_quantity[c, p]
                        contracts_demand_spread[c] -= minimum_quantity[c, p]

    eligible_contracts_for_spread = dict()
    for key, value in contracts_demand_spread.items():
        if value > 0:
            eligible_contracts_for_spread[key] = value
    eligible_contracts_for_spread = sorted(eligible_contracts_for_spread, key=eligible_contracts_for_spread.get, reverse=True)
    '''for contract in eligible_contracts_for_spread:
        possible_routes = get_possible_routes_for_contract(contract, sol_rou, visit_load_unload_port)
        for v in possible_routes:
            contract = choice(eligible_contracts_for_spread)
            if contract not in contracts_on_voyage[v]:
                prod = get_product_for_contract(contract)
                print("Contract", contract, "can be placed on voyage", v)
                if product_capacity[prod][find_vessel(v, y)] - space_used[find_vessel(v, y)] > minimum_quantity[contract, prod]:
                    delta[v, contract] = 1
                    print("CONTRACT", contract, "PLACED ON VOYAGE", v)
                    return delta'''
    return eligible_contracts_for_spread, contracts_demand_spread


def get_product_for_contract(contract):
    for p in products:
        if (contract, p) in contract_demand:
            return p


def get_available_space_on_voyage(delta, y):
    contracts_on_voyage = dict()
    total_load = dict()
    available_space = dict()

    for v in voyages:
        available_space.setdefault(v, dict())
        for p in products:
            available_space[v][p] = vessel_capacity[find_vessel(v, y), p]

    contracts_demand_spread = dict()
    for (c, p), demand in contract_demand.items():
        contracts_demand_spread[c] = demand

    for v in voyages:
        total_load[v] = 0
        contracts_on_voyage[v] = get_contracts_on_voyage(v, delta)
        for c in contracts_on_voyage[v]:
            for p in products:
                if (c, p) in contract_demand:
                    if minimum_pickups[c] <= 1:
                        total_load[v] += contract_demand[c, p]
                        available_space[v][p] -= contract_demand[c, p]
                        contracts_demand_spread[c] -= contract_demand[c, p]
                    else:
                        total_load[v] += minimum_quantity[c, p]
                        contracts_demand_spread[c] -= minimum_quantity[c, p]
                        available_space[v][p] -= minimum_quantity[c, p]
    return available_space


def check_contracts_on_voyages_load(delta, available_space, spread_contracts, contracts_demand_spread, sol_rou,
                                    visit_load_unload_port, space_used, product_capacity, y):
    contract_voyages = dict()
    voyage_contracts = dict()
    extra_load_on_voyage = dict()
    product_move = dict()
    for v in voyages:
        product_move.setdefault(v, dict())
        voyage_contracts.setdefault(v, [])
        extra_load_on_voyage.setdefault(v, dict())
    for c in spread_contracts:
        p = get_product_for_contract(c)
        contract_voyages.setdefault(c, [])
        for v in voyages:
            if (v, c) in delta and delta[v, c] == 1:
                extra_load_on_voyage[v][p] = 0
                contract_voyages[c].append(v)
                voyage_contracts[v].append(c)

    for v in voyages:
        for c in voyage_contracts[v]:
            p = get_product_for_contract(c)
            spread = minimum_pickups[c]
            extra_load_on_voyage[v][p] += contracts_demand_spread[c] / spread
    #print("AVAILABLE SPACE: ", available_space)
    #print("EXTRA LOAD: ", extra_load_on_voyage)

    contracts_from_products_move = dict()
    for v in voyages:
        contracts_from_products_move.setdefault(v, dict())
        for p in products:
            contracts_from_products_move[v].setdefault(p, [])
            if p in extra_load_on_voyage[v] and available_space[v][p] - extra_load_on_voyage[v][p] < 0:
                #print("VOYAGE", v, "DOES NOT HAVE ROOM FOR ALL THE EXTRA LOAD OF PRODUCT ", p)
                product_move[v][p] = available_space[v][p] - extra_load_on_voyage[v][p]
                for c in voyage_contracts[v]:
                    product = get_product_for_contract(c)
                    if product == p:
                        contracts_from_products_move[v][p].append(c)
    '''for v in voyages:
        for p in products:
            if len(contracts_from_products_move[v][p]) > 0:
                for c in contracts_from_products_move[v][p]:
                    if c in evenly_spread_contracts and len(contract_voyages[c]) < maximum_pickups[c]:
                        add_extra_spread(c, delta, sol_rou, visit_load_unload_port, available_space, product_capacity, y)'''


def get_voyages_for_contract(delta, contract):
    voy = []
    for v in voyages:
        if (v, contract) in delta and delta[v, contract] == 1:
            voy.append(v)
    return voy


def get_voyage_times_in_port(t_var):
    time_ports_for_voyage = dict()
    for v in voyages:
        time_ports_for_voyage[v] = dict()
    for value, t in t_var:
        port_voy = t.strip("T").split(",")
        port = int(port_voy[0])
        voyage = int(port_voy[1])
        time_ports_for_voyage[voyage][port] = value
    return time_ports_for_voyage


def get_diff_times_in_port(time_ports_for_voyage):
    diff_times_in_port = dict()
    for i in ports:
        diff_times_in_port.setdefault(i, dict())
        for v in range(1, voyage_count):
            for w in range(v + 1, voyage_count+1):
                time = abs(time_ports_for_voyage[w][i] - time_ports_for_voyage[v][i])
                diff_times_in_port[i][v, w] = time
    return diff_times_in_port


def check_evenly_spread_contracts(diff_times_in_port, delta, sol_rou, visit_load_unload_port, available_space,
                                  product_capacity, y):
    total_time = end_of_planning_horizon
    take_voyage_c = dict()
    for c in evenly_spread_contracts:
        spread = minimum_pickups[c]
        optimal_spread = total_time / spread
        port = load_port[c]
        key = sorted((abs(value-optimal_spread), key) for key, value in diff_times_in_port[port].items())
        possible_routes = get_possible_routes_for_contract(c, sol_rou, visit_load_unload_port)
        best = key[0][0]
        best_spread = deepcopy(spread)
        while key[0][0] > 0.2 and spread < maximum_pickups[c] and spread != 1 and spread < len(possible_routes):
            spread += 1
            optimal_spread = total_time / spread
            key = sorted((abs(value - optimal_spread), key) for key, value in diff_times_in_port[port].items())
            if best > key[0][0]:
                best = key[0][0]
                best_spread = deepcopy(spread)
            voyages_for_contract = get_voyages_for_contract(delta, c)
            #add_extra_spread(c, delta,  sol_rou, visit_load_unload_port, available_space, product_capacity, y)
        take_voyage_c.setdefault(c, [])
        index = 0

        while len(take_voyage_c[c]) < best_spread and index < voyage_count-1 and best_spread > 1:
            voyages = key[index][1]
            index += 1
            if len(take_voyage_c[c]) == 0 and voyages[0] in possible_routes and voyages[1] in possible_routes:
                take_voyage_c[c].append(voyages[0])
                take_voyage_c[c].append(voyages[1])
            elif voyages[0] in possible_routes and voyages[1] in possible_routes:
                if voyages[0] not in take_voyage_c[c] and voyages[1] not in take_voyage_c[c] and len(take_voyage_c[c])+2 <= spread:
                    if max(take_voyage_c) >= voyages[0] and max(take_voyage_c) >= voyages[1]:
                        take_voyage_c[c].append(voyages[0])
                        take_voyage_c[c].append(voyages[1])
                    elif min(take_voyage_c) <= voyages[0] and min(take_voyage_c) <= voyages[1]:
                        take_voyage_c[c].append(voyages[0])
                        take_voyage_c[c].append(voyages[1])
                elif voyages[0] not in take_voyage_c[c]:
                    if min(take_voyage_c[c]) == voyages[1] or max(take_voyage_c[c]) == voyages[1]:
                        take_voyage_c[c].append(voyages[0])
                elif voyages[1] not in take_voyage_c[c]:
                    if min(take_voyage_c[c]) == voyages[0] or max(take_voyage_c[c]) == voyages[0]:
                        take_voyage_c[c].append(voyages[1])
        #print(c, spread, possible_routes, take_voyage_c[c])
        p = get_product_for_contract(c)
        check = True
        for route in take_voyage_c[c]:
            if available_space[find_vessel(route, y)][p] < minimum_quantity[c, p]:
                check = False
        if best_spread > 1 and len(take_voyage_c[c]) > 0 and len(take_voyage_c[c]) >= minimum_pickups[c] and len(take_voyage_c[c]) <= maximum_pickups[c] and check:
            for v in range(1, voyage_count+1):
                if (v, c) in delta and delta[v, c] == 1:
                    if v not in take_voyage_c[c]:
                        #print("REMOVE CONTRACT", c, "FROM VOYAGE", v)
                        delta[v, c] = 0
                elif v in take_voyage_c[c] and ((v,c) not in delta or delta[v, c] == 0):
                    delta[v, c] = 1
                    #print("PLACE CONTRACT", c, "ON VOYAGE", v)
    return delta


def get_sequence_vessels(y, route_and_vessel, solution_routes, sol_routes):
    key = sorted((value, key) for key, value in vessel_available_at.items())
    index = 0
    ind = 1
    rev_index = voyage_count
    y_ = dict()
    for (_, vessel) in key:
        for voyage in voyages:
            if (voyage, vessel) in y:
                route = route_and_vessel[vessel-1][1]
                solution_routes[voyage] = route
                if route != 0:
                    sol_routes[0][ind-1] = route
                    y_[ind, vessel] = 1
                    ind += 1
                else:
                    sol_routes[0][rev_index-1] = route
                    y_[rev_index, vessel] = 1
                    rev_index -= 1
    i = 1
    for r in sol_routes[0]:
        solution_routes[i] = r
        i += 1
    y = y_
    return y, route_and_vessel, solution_routes, sol_routes


def get_load_per_port_for_voyage(delta, y):
    fix = []
    fix_arcs = []
    fix_ports = []
    placeholder = dict()
    placeholder2 = dict()
    contracts_on_voyage = dict()
    load_on_voyage = dict()
    for v in voyages:
        contracts_on_voyage[v] = get_contracts_on_voyage(v, delta)
        load_on_voyage.setdefault(v, dict())
        for i in ports:
            load_on_voyage[v].setdefault(i, dict())
            for p in products:
                load_on_voyage[v][i][p] = 0
            if i > ports[0]:
                load_on_voyage[v][i][p] = load_on_voyage[v][i - 1][p]
            for c in contracts_on_voyage[v]:
                p = get_product_for_contract(c)
                if minimum_pickups[c] > 1:
                    quantity = contract_demand[c, p] / minimum_pickups[c]
                else:
                    quantity = contract_demand[c, p]
                if load_port[c] == i:
                    load_on_voyage[v][i][p] += quantity
                if unload_port[c] == i:
                    load_on_voyage[v][i][p] -= quantity
            for p in products:
                placeholder.setdefault(p, None)
                placeholder2.setdefault(p, None)
                capacity_voyage = 0
                for prod in reversed(products):
                    if p <= prod:
                        capacity_voyage += vessel_capacity[find_vessel(v, y), prod]
                if capacity_voyage - load_on_voyage[v][i][p] < 0:
                    print("CHECK: ", v, i, p)
                    fix.append((v, i, p))
                    if placeholder[p] != None and (v, i-1, p) in fix:
                        placeholder2[p] = i
                    if placeholder[p] == None:
                        placeholder[p] = i
                else:
                    if placeholder[p] != None and placeholder2[p]!= None:
                        fix_arcs.append((v, placeholder[p], placeholder2[p], p))
                        placeholder[p] = None
                        placeholder2[p] = None
                    elif placeholder[p] != None:
                        fix_ports.append((v, placeholder[p], p))
                        placeholder[p] = None
    return fix_arcs, fix_ports


def repair_loads(delta, fix, contracts_time, sol_rou, visit_load_unload_port, available_space, product_capacity, y):
    check = False
    for (v, i, p) in fix:
        contracts_on_voyage = get_contracts_on_voyage(v, delta)
        for c in contracts_on_voyage:
            '''if c not in transit_contracts:
                delta, check = add_extra_spread(c, delta, sol_rou, visit_load_unload_port, available_space, product_capacity, y)
                if check == True:
                    return delta, check'''
    return delta, check




