from heuristic_functions import *
import time
from models.voyage_model_fixed import model_delta_fixed, add_fixed_variables
from copy import deepcopy
import cplex
from random import shuffle, choice
from config import *
from input_for_voyage_model import Svar
MODEL = 1
s_variables = ["S{}".format(c) for c in Svar]
t_variables = ["T{},{}".format(i, v) for (i, v) in Tvar]
t4 = time.time()

success = None
num_success = []

# PART ONE OF HEURISTIC - GET ROUTES (MAYBE Y)
demand_in_port = get_demand_in_port()
routes = create_routes()
visits = visits_for_routes(routes)
route_visits_port = if_route_visits_port(routes, visits)
cost_of_route = cost_of_route(routes, visits)
minimum_visits_port, min_visits_products = find_minimum_visit()
average_capacity = get_average_capacity()
average_capacity_product = get_average_capacity_product()
demand_per_port = get_quantity_per_port(demand_in_port)
visit_load_unload_port = get_visit_load_unload_port(routes, route_visits_port, contracts, contract_ports)
pickups = get_pickups(contracts, minimum_pickups)
#spread_diff = get_spread_max_min()

min_sail_time_contract_route = get_min_sail_time_contract_route(visits, routes)

t1 = time.time()
problem = get_routes_from_model(MODEL, cost_of_route, routes, minimum_visits_port, route_visits_port, ports,
                                voyage_count, vessels, demand_in_port, average_capacity, demand_per_port,
                                average_capacity_product, visit_load_unload_port, pickups, min_sail_time_contract_route)
#print("TIME CREATE ROUTE MODEL: ", time.time()-t1)
t0 = time.time()
problem.solve()
#print("SOLVE ROUTE MODEL: ", time.time()-t0)

sol_routes, solution_routes, sol_vars, sol_vals = get_solutions(MODEL, problem)

print_routes(sol_routes, routes)


# PART 2 OF HEURISTIC - GET Y
t21 = time.time()
y_model = fix_y_model(sol_routes[0], cost_of_route, average_capacity_product, demand_in_port, route_visits_port, min_visits_products, contracts, visit_load_unload_port, min_sail_time_contract_route)
#print("TIME CREATE Y MODEL: ", time.time() - t21)
t22 = time.time()
y_model.solve()
#print("TIME SOLVE Y MODEL: ", time.time() - t22)
y, route_and_vessel = get_y_voyage_vessel(y_model, sol_routes[0])
old_y = deepcopy(y)
old_routes = deepcopy(sol_routes)
old_route_and_vessels = deepcopy(route_and_vessel)

time_r = time_routes(sol_routes, visits, distance_in)

# PART 3 OF HEURISTIC - FIX DELTA
demands, min_pickup = get_demands_and_min_pickup(visit_load_unload_port, sol_routes[0], min_sail_time_contract_route)
product_capacity = get_product_capacity()

t5 = time.time()
voyage_model = model_delta_fixed()
#print("TIME CREATE VOYAGE MODEL: ", time.time()-t5)
t20 = time.time()
voyage_model.write("model" + FILENAME + ".lp")
#print("TIME WRITE VOYAGE MODEL: ", time.time()-t20)

iteration = 1
new_solution = False
best_solution = 1000000000000
first_solution = None
time_to_first_solution = 0
best_solution_problem = None
best_delta = None
best_y = None
model_runs = []
found_delta = 0
if_solution_found = "No solution found ... (yet)"
min_service_level = 1000
stop = False
solution = False
shuffle_it = 0
shuffle_routes = False
solution_it = []
new_route = False
new_y = False

route_it = 1
new_routes_iterations = 1

while time.time()-t4 < TIME and not stop:
    print("--------------------------------------- ITERATION: ", iteration, "-----------------------------------------" + str(best_solution))
    if new_route:  # new_routes_iterations % NEW_ROUTES == 0:
        new_route = False
        new_y = True
        '''if route_it == 2:
            stop = True
            break'''
        print(" *********************************************** NEW ROUTES " + str(route_it) +" ******************************************************** ")
        route_it += 1
        problem.linear_constraints.add(lin_expr=[[sol_vars, sol_vals]],
                                       senses=["L"],
                                       rhs=[int(sum(sol_vals))-1])
        t = time.time()
        problem.solve()
        if problem.solution.get_status() != 103:
            sol_routes, solution_routes, sol_vars, sol_vals = get_solutions(MODEL, problem)
            old_routes = deepcopy(sol_routes)
            time_r = time_routes(sol_routes, visits, distance_in)

    if iteration % NEW_Y == 0 or new_y: # new_routes_iterations % NEW_ROUTES == 0:

        if shuffle_routes and shuffle_it < SHUFFLE and not new_y: #new_routes_iterations % NEW_ROUTES != 0:
            print("#################### SHUFFLE" + str(shuffle_it) + " ####################")
            shuffle(sol_routes[0])
            solution_routes = get_solution_routes(sol_routes)
            #print_routes(sol_routes, routes)
            voyage = 1
            y = dict()
            placed_vessel = []
            for route in sol_routes[0]:
                placed = False
                for vessel, route2 in route_and_vessel:
                    if route2 == route and not placed and vessel not in placed_vessel:
                        y[voyage, vessel] = 1
                        voyage += 1
                        placed = True
                        placed_vessel.append(vessel)
            shuffle_it += 1
            time_r = time_routes(sol_routes, visits, distance_in)
            demands, min_pickup = get_demands_and_min_pickup(visit_load_unload_port, sol_routes[0])
            product_capacity = get_product_capacity()
        else:
            # Get new y-variables (assign vessels to voyages)
            new_routes_iterations += 1
            shuffle_routes = True
            shuffle_it = 0
            sol_routes = deepcopy(old_routes)
            solution_routes = get_solution_routes(sol_routes)
            time_r = time_routes(sol_routes, visits, distance_in)
            if iteration != 1:
                y_model.linear_constraints.add(
                    lin_expr=[[["Y{},{}".format(r - 1, k) for (r, k) in old_y], [1 for (r, k) in old_y]]],
                    senses=["L"],
                    rhs=[int(len(old_y)) - 1])

            if new_y: #new_routes_iterations % NEW_ROUTES == 0:
                # Reset y-model when new routes
                y_model = fix_y_model(sol_routes[0], cost_of_route, average_capacity_product, demand_in_port,
                                      route_visits_port, min_visits_products, contracts, visit_load_unload_port,
                                      min_sail_time_contract_route)
                new_y = False
            y_model.solve()

            if y_model.solution.get_status() != 103:
                y, route_and_vessel = get_y_voyage_vessel(y_model, sol_routes[0])
                old_y = deepcopy(y)
                old_route_and_vessels = deepcopy(route_and_vessel)
                y, route_and_vessel, solution_routes, sol_routes = get_sequence_vessels(y, route_and_vessel, solution_routes, sol_routes)
                time_r = time_routes(sol_routes, visits, distance_in)
                demands, min_pickup = get_demands_and_min_pickup(visit_load_unload_port, sol_routes[0], min_sail_time_contract_route)
                product_capacity = get_product_capacity()
            else:
                new_route = True
                #print("NO NEW Y!")
    #print_routes(sol_routes, routes)
    #print("Y: ", y, route_and_vessel)
    time_read_voy_model_start = time.time()
    model = cplex.Cplex("model" + FILENAME + ".lp")
    t2 = time.time()
    find_sequence = True
    delta, space_used = get_delta(demands, product_capacity, min_pickup, route_visits_port, solution_routes, y, time_r, visit_load_unload_port, sol_routes[0])
    if delta != {}:
        found_delta += 1
        check = True
        avail_space_voyages = get_available_space_on_voyage(delta, y)
        eligible_contracts_for_spread, contracts_demand_spread = get_loads(delta, sol_routes[0], visit_load_unload_port, space_used, product_capacity, y)
        #check_contracts_on_voyages_load(delta, avail_space_voyages, eligible_contracts_for_spread, contracts_demand_spread, sol_routes[0], visit_load_unload_port, space_used, product_capacity, y)
        violating_contracts, delta , contracts_time = check_delta_transit_contracts(delta, y)
        print("VIOLATING CONTRACTS: ", violating_contracts)
        #print("CONTRACTS TIME", contracts_time)
        ex_contracts = []
        check = True
        contracts_switched = []
        while len(violating_contracts) > 0 and check:
            avail_space_voyages = get_available_space_on_voyage(delta, y)
            eligible_contracts_for_spread, contracts_demand_spread = get_loads(delta, sol_routes[0],
                                                                               visit_load_unload_port, space_used,
                                                                               product_capacity, y)
            #check_contracts_on_voyages_load(delta, avail_space_voyages, eligible_contracts_for_spread,
                                            #contracts_demand_spread, sol_routes[0], visit_load_unload_port, space_used, product_capacity, y)
            violating_contracts, delta, contracts_time = check_delta_transit_contracts(delta, y)
            #print("VIOLATING CONTRACTS: ", violating_contracts)
            delta, check, space_used, ex_contracts = repair_transit_time_contract(violating_contracts, delta, sol_routes[0],
                                                                    visit_load_unload_port, space_used,
                                                                    product_capacity, y, avail_space_voyages, ex_contracts)
        print("VIOLATING CONTRACTS: ", violating_contracts)



        print("-------------------------------------- START MODEL RUN ------------------------------------------------")
        #fix_arcs, fix_ports = get_load_per_port_for_voyage(delta, y)
        check = True
        '''print("FIX ARCS", fix_arcs)
        print(("FIX PORTS", fix_ports))'''
        '''while len(fix) > 0 and check:
            fix_arcs, fix_ports = get_load_per_port_for_voyage(delta, y)
            print("FIX", fix)
            avail_space_voyages = get_available_space_on_voyage(delta, y)
            delta, check = repair_loads(delta, fix, contracts_time, sol_routes[0], visit_load_unload_port, avail_space_voyages, product_capacity, y)'''

        model_runs.append(iteration)
        X, DELTA, phi, Z, y = run_model(delta, y, contracts)
        problem2 = add_fixed_variables(model, X, DELTA, phi, Z, y)
        start_voy_model = time.time()
        problem2.solve()
        print("STATUS: ", problem2.solution.get_status_string(problem2.solution.get_status()), problem2.solution.get_status())
        if problem2.solution.get_status() != 3: #== 101 or problem2.solution.get_status() == 102:
            solution_it.append(iteration)
            if first_solution is None:
                first_solution = round(problem2.solution.get_objective_value())
                time_to_first_solution = time.time() - t4
                if_solution_found = "Solution found! (yay)"
            write_solution_to_file(FILENAME, problem2)
            service_level = 0
            max_s = 0
            s_index = None
            for s in s_variables:
                #print(s, problem2.solution.get_values(s))
                service_level += problem2.solution.get_values(s)
                if problem2.solution.get_values(s) > max_s:
                    s_index = s
                    max_s = problem2.solution.get_values(s)
            print("SERVICE LEVEL = ", service_level)
            if service_level <= SERVICE_LEVEL - 0.1:
                print("SUCCESS")
                success = problem2.solution.get_objective_value() - penalty * service_level
                num_success.append((success, service_level))
            if service_level < min_service_level:
                min_service_level = service_level
            if problem2.solution.get_objective_value()-penalty*service_level < best_solution:
                best_solution = round(problem2.solution.get_objective_value() - penalty * service_level)
                best_solution_problem = problem2
                best_delta = deepcopy(delta)
                best_y = deepcopy(y)
            print_solution_voyage_model(problem2, best_solution)
            it = 0
            old_delta = None
            deltas = []
            while service_level > SERVICE_LEVEL and problem2.solution.get_status() != 3 and old_delta != delta \
                    and delta not in deltas:
                deltas.append(delta)
                old_delta = deepcopy(delta)
                it += 1
                s_values = []
                t_values = []
                voy_for_c = dict()

                for t in t_variables:
                    t_values.append((problem2.solution.get_values(t), t))
                time_ports_for_voyage = get_voyage_times_in_port(t_values)
                #print("Time in ports:", time_ports_for_voyage)
                '''for c in evenly_spread_contracts:
                    voy_for_c[c] = get_voyages_for_contract(delta, c)
                diff_times_in_port = get_diff_times_in_port(time_ports_for_voyage)
                delta = check_evenly_spread_contracts(diff_times_in_port, delta, sol_routes[0], visit_load_unload_port,
                                              avail_space_voyages, product_capacity, y)'''

                for s in s_variables:
                    s_values.append((problem2.solution.get_values(s), s))
                    #print(s, problem2.solution.get_values(s))
                    if problem2.solution.get_values(s) > max_s:
                        s_index = s
                        max_s = problem2.solution.get_values(s)
                contract_spread = int(s_index.strip("S"))
                avail_space_voyages = get_available_space_on_voyage(delta, y)
                for val, s in reversed(sorted(s_values)):
                    if val > S_LEVEL_SEARCH:
                        delta, switched = search_delta(s_variables, s, delta, sol_routes[0], visit_load_unload_port, avail_space_voyages)
                    if switched:
                        break
                violating_contracts, delta, contracts_time = check_delta_transit_contracts(delta, y)
                ex_contracts = []
                check = True
                contracts_switched = []
                while len(violating_contracts) > 0 and check:
                    avail_space_voyages = get_available_space_on_voyage(delta, y)
                    eligible_contracts_for_spread, contracts_demand_spread = get_loads(delta, sol_routes[0],
                                                                                       visit_load_unload_port,
                                                                                       space_used,
                                                                                       product_capacity, y)
                    check_contracts_on_voyages_load(delta, avail_space_voyages, eligible_contracts_for_spread,
                                                    contracts_demand_spread, sol_routes[0], visit_load_unload_port,
                                                    space_used, product_capacity, y)
                    violating_contracts, delta, contracts_time = check_delta_transit_contracts(delta, y)
                    # print("VIOLATING CONTRACTS: ", violating_contracts)
                    delta, check, space_used, ex_contracts = repair_transit_time_contract(violating_contracts, delta,
                                                                                          sol_routes[0],
                                                                                          visit_load_unload_port,
                                                                                          space_used,
                                                                                          product_capacity, y,
                                                                                          avail_space_voyages,
                                                                                          ex_contracts)
                print("VIOLATING CONTRACTS: ", violating_contracts)
                print(
                    "-------------------------------------- START MODEL RUN AFTER SEARCHing", it, "------------------------------------------------")
                problem2 = cplex.Cplex("model" + FILENAME + ".lp")
                problem2.linear_constraints.add(lin_expr=[[["S{}".format(c) for c in evenly_spread_contracts],
                                              [1.0]*(len(evenly_spread_contracts))]],
                                   senses=["L"],
                                   rhs=[service_level])
                #model_runs.append(iteration)
                X, DELTA, phi, Z, y = run_model(delta, y, contracts)
                problem2 = add_fixed_variables(model, X, DELTA, phi, Z, y)
                start_voy_model = time.time()
                problem2.solve()
                print("STATUS: ", problem2.solution.get_status_string(problem2.solution.get_status()),
                      problem2.solution.get_status())

                if problem2.solution.get_status() != 3:
                    print_solution_voyage_model(problem2, best_solution)
                    write_solution_to_file(FILENAME, problem2)
                    service_level = 0
                    max_s = 0
                    s_index = None
                    for s in s_variables:
                        service_level += problem2.solution.get_values(s)
                        if problem2.solution.get_values(s) > max_s:
                            s_index = s
                            max_s = problem2.solution.get_values(s)
                    if problem2.solution.get_objective_value() - penalty * service_level < best_solution:
                        best_solution = round(problem2.solution.get_objective_value() - penalty * service_level)
                        best_solution_problem = problem2
                        best_delta = deepcopy(delta)
                        best_y = deepcopy(y)
                    print("SERVICE LEVEL = ", service_level)
                    if service_level <= SERVICE_LEVEL - 0.1:
                        #print("SUCCESS")
                        success = problem2.solution.get_objective_value() - penalty * service_level
                        num_success.append((success, service_level))
                    if service_level < min_service_level:
                        min_service_level = service_level
                        #num_success.append((min_service_level, problem2.solution.get_objective_value()-penalty*(min_service_level)))'''

            print("------------------------------------ END MODEL RUN", it,"------------------------------------------")


            if service_level <= SERVICE_LEVEL + 0.1:
                print("SUCCESS")
                success = problem2.solution.get_objective_value() - penalty * service_level
                num_success.append((success, service_level))
        else:
            shuffle_routes = True
            '''excluding = []
            spread_contract = 1
            while problem2.solution.get_status() == 3 and spread_contract != 0:
                placed = False
                while not placed and spread_contract != 0:
                    spread_contract = get_max_diff_spread(excluding, spread_diff)
                    if spread_contract != 0:
                        possible_routes = get_possible_routes_for_contract(spread_contract, sol_routes[0], visit_load_unload_port)
                        print("TESTING EXTRA VOYAGE FOR CONTRACT", spread_contract)
                        for p in products:
                            if (spread_contract, p) in contract_demand:
                                prod = p
                        for v in possible_routes:
                            if (v, spread_contract) not in delta and not placed and ((product_capacity[prod][find_vessel(v, y)] - space_used[find_vessel(v, y)]) > minimum_quantity[spread_contract, prod]):
                                delta[v, spread_contract] = 1
                                print("PLACED ON VOYAGE: ", v)
                                placed = True
                        excluding.append(spread_contract)
                        X, DELTA, phi, Z, y = run_model(delta, y, contracts)
                        problem2 = add_fixed_variables(model, X, DELTA, phi, Z, y)
                        start_voy_model = time.time()
                        problem2.solve()
                        print("TIME SOLVE VOY MODEL: ", time.time() - start_voy_model)
                        print("STATUS: ", problem2.solution.get_status_string(problem2.solution.get_status()),
                              problem2.solution.get_status())
                        # print("DELTA: ", delta)


                        if problem2.solution.get_status() != 3:  # == 101 or problem2.solution.get_status() == 102:
                            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!SOLUTION!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                            if first_solution is None:
                                first_solution = round(problem2.solution.get_objective_value())
                                time_to_first_solution = time.time() - t4
                                if_solution_found = "Solution found! (yay)"

                            if problem2.solution.get_objective_value() < best_solution:
                                best_solution = round(problem2.solution.get_objective_value())
                                best_solution_problem = problem2
                                best_delta = deepcopy(delta)
                                best_y = deepcopy(y)

                            print_solution_voyage_model(problem2, best_solution)
                            write_solution_to_file(FILENAME, problem2)
                            service_level = 0
                            max_s = 0
                            s_index = None
                            for s in s_variables:
                                service_level += problem2.solution.get_values(s)
                                if problem2.solution.get_values(s) > max_s:
                                    s_index = s
                                    max_s = problem2.solution.get_values(s)
                            print("SERVICE LEVEL = ", service_level)
                            if service_level < min_service_level:
                                min_service_level = service_level
                                num_success.append((min_service_level, problem2.solution.get_objective_value()))
                            it = 0
                            old_delta = None
                            while service_level > SERVICE_LEVEL and problem2.solution.get_status() != 3 and old_delta != delta:
                                old_delta = deepcopy(delta)
                                it += 1
                                s_values = []
                                for s in s_variables:
                                    s_values.append((problem2.solution.get_values(s), s))
                                    if problem2.solution.get_values(s) > max_s:
                                        s_index = s
                                        max_s = problem2.solution.get_values(s)

                                for val, s in reversed(sorted(s_values)):
                                    if val > S_LEVEL_SEARCH:
                                        delta, switched = search_delta(s_variables, s, delta, sol_routes[0], visit_load_unload_port)
                                    if switched:
                                        break
                                print(
                                    "-------------------------------------- START MODEL RUN AFTER SEARCH", it,
                                    "------------------------------------------------")
                                model = cplex.Cplex("model" + FILENAME + ".lp")
                                model_runs.append(iteration)
                                X, DELTA, phi, Z, y = run_model(delta, y, contracts)
                                problem2 = add_fixed_variables(model, X, DELTA, phi, Z, y)
                                start_voy_model = time.time()
                                problem2.solve()

                                if problem2.solution.get_status() != 3:

                                    if first_solution is None:
                                        first_solution = round(problem2.solution.get_objective_value())
                                        time_to_first_solution = time.time() - t4
                                        if_solution_found = "Solution found! (yay)"

                                    if problem2.solution.get_objective_value() < best_solution:
                                        best_solution = round(problem2.solution.get_objective_value())
                                        best_solution_problem = problem2
                                        best_delta = deepcopy(delta)
                                        best_y = deepcopy(y)

                                    print_solution_voyage_model(problem2, best_solution)
                                    write_solution_to_file(FILENAME, problem2)
                                    service_level = 0
                                    max_s = 0
                                    s_index = None
                                    for s in s_variables:
                                        service_level += problem2.solution.get_values(s)
                                        if problem2.solution.get_values(s) > max_s:
                                            s_index = s
                                            max_s = problem2.solution.get_values(s)
                                    print("SERVICE LEVEL = ", service_level)
                                    if service_level < min_service_level:
                                        min_service_level = service_level
                                        num_success.append((min_service_level, problem2.solution.get_objective_value()))

                            print("------------------------------------ END MODEL RUN", it,
                                  "------------------------------------------")

                        if service_level <= SERVICE_LEVEL:
                            print("SUCCESS")
                            success = problem2.solution.get_objective_value()
                            num_success.append((success, service_level))'''

    else:
        shuffle_routes = False
    iteration += 1



run_time = time.time() - t4

with open('results/delta_y/' + FILENAME, 'a') as f:
    f.write(str(FILENAME) + "\n\n")
    f.write(str(best_delta) + "\n\n")
    f.write(str(best_y) + "\n")

print("---- MODEL ", MODEL, " ----")
print("FILENAME: ", FILENAME)
print("TOTAL TIME: ", run_time)
print("BEST SOLUTION: ", best_solution)
print("TIME FIRST SOLUTION: ", time_to_first_solution)
print("MODEL RUNS: ", model_runs)
print("---------------------------------------------")
print("SUCCESS? ", sorted(num_success))
print("MINIMUM SERVICE LEVEL: ", min_service_level)
print("GOAL SERVICE LEVEL: ", SERVICE_LEVEL)

print("SOLUTIONS ITERATIONS: ", solution_it)
if len(solution_it) != 0:
    print("PROSENT GYLDIG AV TESTET:", "gyldige løsninger:", len(solution_it)," - Kjør:", len(model_runs), "- prosent: ",(len(solution_it)/len(model_runs))*100)

if first_solution:
    with open('results/parameters/' + FILENAME, 'a') as f:
        f.write(str(FILENAME) + "\t")
        f.write("NEW_Y: " + str(NEW_Y) + "\t")
        f.write("NEW_ROUTES: " + str(NEW_ROUTES) + "\t")
        f.write("BUFFER: " + str(BUFFER) + "\t")
        f.write("BUFFER1: " + str(BUFFER1) + "\t")
        f.write("SERVICE LEVEL: " + str(SERVICE_LEVEL) + "\t")
        f.write("TIME: " + str(TIME) + "\t")
        f.write("BEST SOLUTION: " + str(best_solution) + "\t")
        f.write("Max service level : " + str(service_level) + "\t")
        f.write("Min service level : " + str(min_service_level) + "\n")



