import numpy as np
from math import floor
from shared_for_input import *
import time

t0 = time.time()

voyages = [i for i in range(1, voyage_count+1)]
contracts = [i for i in range(1, contract_count+1)]
ports = [i for i in range(first_port, last_port+1)]
products = [i for i in range(1, product_count+1)]
velocities = [i for i in range(1, velocities_count+1)]
nodes = [i for i in range(1, last_node+1)]

load_port = {}
unload_port = {}
for c in contracts:
    load_port[c] = contract_ports[c][0]
    unload_port[c] = contract_ports[c][1]

arcs = {}
for k in vessels:
    for p in ports:
        arcs[origin[k], p, k] = 1
        arcs[p, destination[k], k] = 1
        for q in range(p+1, last_port+1):
            arcs[p, q, k] = 1
    arcs[origin[k], destination[k], k] = 1


arc = {}
for i in nodes:
    for j in nodes:
        for k in vessels:
            if (i, j, k) in arcs:
                arc[i, j] = 1


distance = {}
for i in ports:
    for j in ports:
        if i < j:
            distance[i, j] = distance_in[i, j]

for k in vessels:
    for j in ports:
        if vessel_origin[k] != j:
            distance[k, j] = distance_in[vessel_origin[k], j]
        else:
            distance[k, j] = 0
        distance[j, destination[k]] = 0
    distance[k, destination[k]] = 0

sailing_time_2 = {}
for i in nodes:
    for j in nodes:
        for k in vessels:
            for s in velocities:
                if (i, j) in distance:
                    sailing_time_2[i, j, k, s] = distance[i, j] * sailing_time[k, s]


leg_sailing_cost = {}
for i in nodes:
    for j in nodes:
        for k in vessels:
            for s in velocities:
                if (i, j) in distance and distance[i, j] >= 0.5:
                    leg_sailing_cost[i, j, k, s] = distance[i, j] * sailing_cost[k, s] + sailing_time_2[i, j, k, s] * charter_cost[k]
                else:
                    leg_sailing_cost[i, j, k, s] = 0

# Subsets
contracts_L = {}
contracts_U = {}
products_S = {}
vessel_nodes = {}
transit_contracts = []

for c in contracts:
    if contract_transit_time[c] >= 0.5:
        transit_contracts.append(c)

for i in ports:
    contracts_L.setdefault(i, [])
    contracts_U.setdefault(i, [])
    for c in contracts:
        if load_port[c] == i:
            contracts_L[i].append(c)
        elif unload_port[c] == i:
            contracts_U[i].append(c)

for p in products:
    products_S.setdefault(p, [])
    for q in products:
        if q > p:
            products_S[p].append(q)

for k in vessels:
    vessel_nodes.setdefault(k, [])
    for p in ports:
        vessel_nodes[k].append(p)
    vessel_nodes[k].append(k)
    vessel_nodes[k].append(k + port_count + vessel_count)

# --------------------------- VARIABLES ----------------------------------
Xvar = []
Lvar = []
for i in nodes:
    for j in nodes:
        for v in voyages:
            if (i, j) in arc:
                Xvar.append((i, j, v))
                for p in products:
                    Lvar.append((i, j, v, p))

Wvar = []
for i in nodes:
    for j in nodes:
        for v in voyages:
            for k in vessels:
                for s in velocities:
                    if (i, j, k) in arcs:
                        Wvar.append((i, j, v, k, s))

Yvar = []
for v in voyages:
    for k in vessels:
        Yvar.append((v, k))

Qvar = []
Dvar = []
for c in contracts:
    for v in voyages:
        for p in products:
            if (c, p) in contract_demand and contract_demand[c, p] > 0:
                Qvar.append((v, c, p))
        Dvar.append((v, c))

Tvar = []
Zvar = []
for i in nodes:
    for v in voyages:
        Tvar.append((i, v))

for c in evenly_spread_contracts:
    for v in voyages:
        for w in voyages:
            if w > v:
                Zvar.append((v, w, c))


Svar = []
Avar = []
for c in evenly_spread_contracts:
    Svar.append(c)
    for n in range(minimum_pickups[c], maximum_pickups[c]+1):
        Avar.append((c, n))

WTvar = []
for k in vessels:
    WTvar.append(k)

Ovar = []
for c in evenly_spread_contracts:
    for n in range(minimum_pickups[c], maximum_pickups[c]+1):
        Ovar.append((n, c))

# --------------------------------------------------------------------
# Max capacity
max_capacity = {}
for p in products:
    max_capacity[p] = 0
    for k in vessels:
        if max_capacity[p] <= vessel_capacity[k, p]:
            max_capacity[p] = vessel_capacity[k, p]


maximum_time_between_ports = {}
for i in nodes:
    for j in nodes:
        eksisterer = False
        maximumofall = 0
        max_port_time2 = 0
        for k in vessels:
            if i in node_time and max_port_time2 < node_time[i]:
                max_port_time2 = node_time[i]
            if (i, j, k, 1) in sailing_time_2:
                eksisterer = True
                if maximumofall < sailing_time_2[i, j, k, 1]:
                    maximumofall = sailing_time_2[i, j, k, 1]
        if eksisterer:
            for v in voyages:
                maximum_time_between_ports[i, j, v] = maximumofall + max_port_time2

'''max_cost_to_port1 = 0
for k in vessels:
    if sailing_time_2[k, 1, k, 1] > max_cost_to_port1:
        max_cost_to_port1 = sailing_time_2[k, 1, k, 1]'''

maximum_time_between_ports_without_voyages = {}
for i in ports:
    for j in ports:
        maximum_time_between_ports_without_voyages[i, j] = 0
        for v in voyages:
            if i < j and maximum_time_between_ports_without_voyages[i, j] < maximum_time_between_ports[i, j, v]:
                maximum_time_between_ports_without_voyages[i, j] = maximum_time_between_ports[i, j, v]

late_time = 0
for k in vessels:
    if late_time < vessel_available_at[k]:
        late_time = vessel_available_at[k]


# -----------------------------------------------------------------------------------
# Big M
# -----------------------------------------------------------------------------------

big_D = {}
for c in contracts:
    iter_big_D = load_port[c]
    big_D[c] = 0
    while iter_big_D < unload_port[c]:
        big_D[c] += maximum_time_between_ports_without_voyages[iter_big_D, (iter_big_D+1)]
        iter_big_D += 1
    big_D[c] += big_D[c] + late_time

time_max = 0
for c in contracts:
    if time_max <= big_D[c]:
        time_max = big_D[c]
time_max += end_of_planning_horizon - late_time

big_E = {}
for c in evenly_spread_contracts:
    big_E[c] = big_D[c] + end_of_planning_horizon
    big_E[c] = big_E[c] + (end_of_planning_horizon/minimum_pickups[c])

max_time_load_ship = 0
n_vessels_real = vessel_count
max_waiting = floor(30.0/(n_vessels_real-1))


for v in vessels:
    temp_time_load_ship = 0
    for p in products:
        temp_time_load_ship += vessel_capacity[v, p] * handling_times[p]
    if temp_time_load_ship >= max_time_load_ship:
        max_time_load_ship = temp_time_load_ship

big_T = {}

for c in transit_contracts:
    iter_big_T = load_port[c]
    big_T[c] = 0
    while iter_big_T < unload_port[c]:
        big_T[c] += maximum_time_between_ports_without_voyages[iter_big_T, (iter_big_T+1)]
        iter_big_T += 1
    big_T[c] += max_time_load_ship + max_waiting


big_S = {}
for k in vessels:
    big_S[first_port, k] = vessel_available_at[k] + max_waiting + sailing_time_2[origin[k], first_port, k, 1] + max_time_load_ship + node_time[first_port]
    for i in ports:
        if i != first_port:
            big_S[i, k] = big_S[i-1, k] + maximum_time_between_ports_without_voyages[(i-1), i] + node_time[i]

big_L = {}
time_max = 0
iter_big_D = first_port
while iter_big_D < last_port:
    time_max += maximum_time_between_ports_without_voyages[iter_big_D, (iter_big_D+1)]
    iter_big_D += 1

time_max += max_time_load_ship * 2 + max_waiting
for k in vessels:
    big_L[k] = time_max + late_time

big_E = {}
for c in contracts:
    big_D[c] = 0
    iter_big_D = load_port[c]
    while iter_big_D < unload_port[c]:
        big_D[c] += maximum_time_between_ports_without_voyages[iter_big_D, (iter_big_D+1)]
        iter_big_D += 1

for c in evenly_spread_contracts:
    big_E[c] = 0
    big_E[c] = big_D[c] + 2 * max_time_load_ship + max_waiting + 2 * end_of_planning_horizon

vessel_cost = dict()
max_cost = dict()

for v in velocities:
    max_cost[v] = 0
    for k in vessels:
        if sailing_cost[k, v] > max_cost[v]:
            max_cost[v] = sailing_cost[k, v]

for k in vessels:
    vessel_cost[k] = sailing_cost[k, 1]/max_cost[1]

t1 = time.time()
print("TIME INPUT: ", t1-t0)

most_space = dict()
for p in products:
    most_space.setdefault(p, [])
    for v in voyages:
        most_space[p].append(vessel_capacity[v, p])

    most_space[p] = sorted(most_space[p], reverse=True)


