from input_for_voyage_model import *
import cplex
import time

def model_delta_fixed(X, DELTA, phi, Z, y):
    t1 = time.time()
    problem = cplex.Cplex()
    problem.objective.set_sense(problem.objective.sense.minimize)
    binary = problem.variables.type.binary

    # -----------------------------------------------------------------------------------
    # Parameters
    # -----------------------------------------------------------------------------------
    service_level = 100000

    # set gap
    #problem.parameters.mip.tolerances.mipgap.set(0.01)

    # -----------------------------------------------------------------------------------
    # Variables
    # -----------------------------------------------------------------------------------
    x_variables = ["X{},{},{}".format(i, j, v) for (i, j, v) in Xvar]
    w_variables = ["W{},{},{},{},{}".format(i, j, v, k, s) for (i, j, v, k, s) in Wvar]
    l_variables = ["L{},{},{},{}".format(i, j, v, p) for (i, j, v, p) in Lvar]
    y_variables = ["Y{},{}".format(v, k) for (v, k) in Yvar]
    q_variables = ["Q{},{},{}".format(v, c, p) for (v, c, p) in Qvar]
    d_variables = ["D{},{}".format(v, c) for (v, c) in Dvar]
    t_variables = ["T{},{}".format(i, v) for (i, v) in Tvar]
    z_variables = ["Z{},{},{}".format(v, w, c) for (v, w, c) in Zvar]
    s_variables = ["S{}".format(c) for c in Svar]
    wt_variables = ["WT{}".format(k) for k in WTvar]

    o_variables = ["O{},{}".format(n, c) for (n, c) in Ovar]

    problem.variables.add(names=x_variables,
                          types=[binary for i in range(len(x_variables))])
    problem.variables.add(names=l_variables)
    problem.variables.add(names=y_variables,
                          types=[binary for i in range(len(Yvar))])
    problem.variables.add(names=q_variables)
    problem.variables.add(names=t_variables)
    '''problem.variables.add(names=w_variables,
                          lb=[0]*len(Wvar),
                          ub=[1]*len(Wvar))'''
    problem.variables.add(names=d_variables,
                          types=[binary for i in range(len(Dvar))])

    problem.variables.add(names=z_variables,
                          types=[binary for i in range(len(Zvar))])
    problem.variables.add(names=s_variables,
                          lb=[0 for c in evenly_spread_contracts])
    '''problem.variables.add(names=wt_variables)'''
    problem.variables.add(names=o_variables,
                          types=[binary for i in range(len(Ovar))])

    # -----------------------------------------------------------------------------------
    # Objective Function
    # -----------------------------------------------------------------------------------
    problem.variables.add(obj=[leg_sailing_cost[i, j, k, s] + charter_cost[k] * (node_time[i] if i in node_time else 0) for (i, j, v,  k, s) in Wvar],
                          names=["W{},{},{},{},{}".format(i, j, v, k, s) for (i, j, v, k, s) in Wvar])
    problem.variables.add(obj=[port_cost[i] if i in ports else 0 for i in ports for j in nodes for v in voyages if (i, j, v) in Xvar],
                          names=["X{},{},{}".format(i, j, v) for i in ports for j in nodes for v in voyages if (i, j, v) in Xvar],
                          types=[binary for i in ports for j in nodes for v in voyages if (i, j, v) in Xvar])
    problem.variables.add(obj=[charter_cost[k] for k in vessels],
                          names=["WT{}".format(k) for k in vessels])

    name2idx = {n: j for j, n in enumerate(problem.variables.get_names())}

    # -----------------------------------------------------------------------------------
    #  Vessel Constraints
    # -----------------------------------------------------------------------------------
    # Constraints 5
    '''problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(i, j, v)] +
                                              ["W{},{},{},{},{}".format(i, j, v, k, s)
                                               for s in velocities for k in vessels if (i, j, k) in arcs],
                                              [-1.0] +
                                              [1.0 for s in velocities for k in vessels if (i, j, k) in arcs]] for (i, j, v) in Xvar],
                                   senses=["E"]*len(Xvar),
                                   rhs=[0]*len(Xvar))

    # Constraints 8
    problem.linear_constraints.add(lin_expr=[[["W{},{},{},{},{}".format(i, j, v, k, s)
                                                for s in velocities if (i, j, k) in arcs] +
                                                 ["Y{},{}".format(v, k)],
                                              [1 for s in velocities if (i, j, k) in arcs] +
                                              [-1]] for (i, j, v) in Xvar for k in vessels],
                                   senses=["L"]*vessel_count*len(Xvar),
                                   rhs=[0]*vessel_count*len(Xvar))'''

    # Constraints 5
    for (i, j, v) in Xvar:
        ind = []
        val = []
        ind.append(name2idx["X{},{},{}".format(i, j, v)])
        val.append(-1)
        for k in vessels:
            for s in velocities:
                if (i, j, k) in arcs:
                    ind.append(name2idx["W{},{},{},{},{}".format(i, j, v, k, s)])
                    val.append(1)
        problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                  val=val)],
                                       senses=["E"],
                                       rhs=[0])

    # Constraints 8
    for (i, j, v) in Xvar:
        for k in vessels:
            ind = []
            val = []
            if (i, j, k) in arcs:
                for s in velocities:
                    ind.append(name2idx["W{},{},{},{},{}".format(i, j, v, k, s)])
                    val.append(1)
            ind.append(name2idx["Y{},{}".format(v, k)])
            val.append(-1)
            problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                      val=val)],
                                           senses=["L"],
                                           rhs=[0])

    # -----------------------------------------------------------------------------------
    # Capacity Constraints
    # -----------------------------------------------------------------------------------
    # Constraints 11
    '''problem.linear_constraints.add(lin_expr=[[["L{},{},{},{}".format(i, j, v, p)] +
                                                 ["Y{},{}".format(v, k) for k in vessels] +
                                                 ["L{},{},{},{}".format(i, j, v, q) for q in products_S[p]],
                                              [-1] +
                                              [vessel_capacity[k, p] for k in vessels] +
                                              [-1]*len(products_S[p])] for p in products for (i, j) in arc for v in voyages],
                                   senses=["G"]*product_count*len(arc)*voyage_count,
                                   rhs=[0]*product_count*len(arc)*voyage_count)

    # Constraints 12
    problem.linear_constraints.add(lin_expr=[[["L{},{},{},{}".format(i, j, v, p)] +
                                                ["X{},{},{}".format(i, j, v)],
                                              [1] +
                                              [(-1)*max_capacity[p]]] for i in ports for j in nodes for v in voyages for p in products if (i, j, v) in Xvar],
                                   senses=["L" for i in ports for j in nodes for v in voyages for p in products if (i, j, v) in Xvar],
                                   rhs=[0 for i in ports for j in nodes for v in voyages for p in products if (i, j, v) in Xvar])

    # Constraints 13
    problem.linear_constraints.add(lin_expr=[[["L{},{},{},{}".format(j, i, v, p) for j in nodes if (j, i, v, p) in Lvar] +
                                                ["Q{},{},{}".format(v, c, p) for c in (contracts_L[i] if i in ports else []) if (v, c, p) in Qvar] +
                                                 ["Q{},{},{}".format(v, c, p) for c in (contracts_U[i] if i in ports else []) if (v, c, p) in Qvar] +
                                                 ["L{},{},{},{}".format(i, j, v, p) for j in nodes if (i, j, v, p) in Lvar],
                                              [1 for j in nodes if (j, i, v, p) in Lvar] +
                                              [1 for c in (contracts_L[i] if i in ports else []) if (v, c, p) in Qvar] +
                                              [-1 for c in (contracts_U[i] if i in ports else []) if (v, c, p) in Qvar] +
                                              [-1 for j in nodes if (i, j, v, p) in Lvar]] for i in nodes for v in vessels for p in products],
                                   senses=["E"]*product_count*last_node*vessel_count,
                                   rhs=[0]*product_count*last_node*vessel_count)'''

    # Constraints 11
    for p in products:
        for (i, j) in arc:
            for v in voyages:
                ind = []
                val = []
                ind.append(name2idx["L{},{},{},{}".format(i, j, v, p)])
                val.append(-1)
                for k in vessels:
                    ind.append(name2idx["Y{},{}".format(v, k)])
                    val.append(vessel_capacity[k, p])
                for q in products_S[p]:
                    ind.append(name2idx["L{},{},{},{}".format(i, j, v, q)])
                    val.append(-1)
                problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                          val=val)],
                                               senses=["G"],
                                               rhs=[0])

    # Constraints 12
    for i in ports:
        for j in nodes:
            for v in voyages:
                for p in products:
                    if (i, j, v) in Xvar:
                        ind = []
                        val = []
                        ind.append(name2idx["L{},{},{},{}".format(i, j, v, p)])
                        val.append(1)
                        ind.append(name2idx["X{},{},{}".format(i, j, v)])
                        val.append((-1) * max_capacity[p])
                        problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                                  val=val)],
                                                       senses=["L"],
                                                       rhs=[0])

    # Constraints 13
    for i in nodes:
        for v in vessels:
            for p in products:
                ind = []
                val = []
                for j in nodes:
                    if (j, i, v, p) in Lvar:
                        ind.append(name2idx["L{},{},{},{}".format(j, i, v, p)])
                        val.append(1)
                    if (i, j, v, p) in Lvar:
                        ind.append(name2idx["L{},{},{},{}".format(i, j, v, p)])
                        val.append(-1)
                if i in ports:
                    for c in contracts_L[i]:
                        if (v, c, p) in Qvar:
                            ind.append(name2idx["Q{},{},{}".format(v, c, p)])
                            val.append(1)
                    for c in contracts_U[i]:
                        if (v, c, p) in Qvar:
                            ind.append(name2idx["Q{},{},{}".format(v, c, p)])
                            val.append(-1)
                problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                          val=val)],
                                               senses=["E"],
                                               rhs=[0])

    # -----------------------------------------------------------------------------------
    # Loading Constraints
    # -----------------------------------------------------------------------------------
    # Constraints 18
    '''problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)] +
                                              ["Q{},{},{}".format(v, c, p)],
                                              [(-1)*maximum_quantity[c, p]] +
                                              [1]] for (v, c, p) in Qvar if (c, p) in maximum_quantity],
                                   senses=["L" for (v, c, p) in Qvar if (c, p) in maximum_quantity],
                                   rhs=[0 for (v, c, p) in Qvar if (c, p) in maximum_quantity])

    problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)] +
                                              ["Q{},{},{}".format(v, c, p)],
                                              [(-1)*minimum_quantity[c, p]] +
                                              [1]] for (v, c, p) in Qvar if (c, p) in minimum_quantity],
                                   senses=["G" for (v, c, p) in Qvar if (c, p) in minimum_quantity],
                                   rhs=[0 for (v, c, p) in Qvar if (c, p) in minimum_quantity])



    # Constraints 19 NB: E i xpress
    problem.linear_constraints.add(lin_expr=[[["Q{},{},{}".format(v, c, p) for v in voyages],
                                              [1 for v in voyages]] for c in contracts for p in products if (c, p) in contract_demand and contract_demand[c, p] > 0],
                                   senses=["E" for c in contracts for p in products if (c, p) in contract_demand and contract_demand[c, p] > 0],
                                   rhs=[contract_demand[c, p] for c in contracts for p in products if (c, p) in contract_demand and contract_demand[c, p] > 0])'''

    # Constraints 18
    problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["D{},{}".format(v, c)]] +
                                                                  [name2idx["Q{},{},{}".format(v, c, p)]],
                                                              val=[(-1) * maximum_quantity[c, p]] +
                                                                  [1]) for (v, c, p) in Qvar if
                                             (c, p) in maximum_quantity],
                                   senses=["L" for (v, c, p) in Qvar if (c, p) in maximum_quantity],
                                   rhs=[0 for (v, c, p) in Qvar if (c, p) in maximum_quantity])

    problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["D{},{}".format(v, c)]] +
                                                                  [name2idx["Q{},{},{}".format(v, c, p)]],
                                                              val=[(-1) * minimum_quantity[c, p]] +
                                                                  [1]) for (v, c, p) in Qvar if
                                             (c, p) in minimum_quantity],
                                   senses=["G" for (v, c, p) in Qvar if (c, p) in minimum_quantity],
                                   rhs=[0 for (v, c, p) in Qvar if (c, p) in minimum_quantity])

    # Constraints 19 NB: E i xpress
    problem.linear_constraints.add(
        lin_expr=[cplex.SparsePair(ind=[name2idx["Q{},{},{}".format(v, c, p)] for v in voyages],
                                   val=[1 for v in voyages]) for c in contracts for p in products if
                  (c, p) in contract_demand and contract_demand[c, p] > 0],
        senses=["E" for c in contracts for p in products if (c, p) in contract_demand and contract_demand[c, p] > 0],
        rhs=[contract_demand[c, p] for c in contracts for p in products if
             (c, p) in contract_demand and contract_demand[c, p] > 0])

    # -----------------------------------------------------------------------------------
    # Time Constraints
    # -----------------------------------------------------------------------------------
    # Constraints 20
    '''problem.linear_constraints.add(lin_expr=[[["T{},{}".format(origin[k], v)] +
                                                 ["Y{},{}".format(v, k)],
                                              [-1] +
                                              [vessel_available_at[k]]] for v in voyages for k in vessels],
                                   senses=["E"]*voyage_count*vessel_count,
                                   rhs=[0]*voyage_count*vessel_count)

    # Constraints 21
    problem.linear_constraints.add(lin_expr=[[["T{},{}".format(i, v)] +
                                              ["W{},{},{},{},{}".format(i, j, v, k, s) for s in velocities for k in vessels if (i, j, v, k, s) in Wvar] +
                                              ["X{},{},{}".format(i, j, v)] +
                                              ["Q{},{},{}".format(v, c, p) for c in (contracts_U[i]+contracts_L[i] if i in ports else []) for p in products if (v, c, p) in Qvar] +
                                              ["T{},{}".format(j, v)],
                                              [1.0] +
                                              [sailing_time_2[i, j, k, s] for s in velocities for k in vessels if (i, j, v, k, s) in Wvar] +
                                              [node_time[i] if i in ports else 0] +
                                              [handling_times[p] for c in (contracts_U[i]+contracts_L[i] if i in ports else [])for p in products if (v, c, p) in Qvar] +
                                              [-1.0]] for (i, j) in arc for v in voyages],
                                   senses=["L"]*len(arc)*voyage_count,
                                   rhs=[0]*len(arc)*voyage_count)

    # Constraints 22
    problem.linear_constraints.add(lin_expr=[[["T{},{}".format(load_port[c], v)] +
                                              ["T{},{}".format(unload_port[c], v)] +
                                              ["D{},{}".format(v, c)],
                                              [1.0] +
                                              [-1.0] +
                                              [(-1)*big_T[c]]] for v in voyages for c in transit_contracts],
                                   senses=["G"]*voyage_count*len(transit_contracts),
                                   rhs=[(-1)*(big_T[c] + contract_transit_time[c]) for v in voyages for c in transit_contracts])

    # Constraints 23
    problem.linear_constraints.add(lin_expr=[[["T{},{}".format(j, v)] +
                                              ["X{},{},{}".format(origin[k], j, v)],
                                              [1.0] +
                                              [big_S[j, k]]] for j in ports for k in vessels for v in voyages],
                                   senses=["L"]*port_count*vessel_count*voyage_count,
                                   rhs=[(big_S[j, k] + end_of_planning_horizon) for j in ports for k in vessels for v in voyages])

    # Constraints 24
    problem.linear_constraints.add(lin_expr=[[["WT{}".format(k)] +
                                              ["T{},{}".format(destination[k], v)] +
                                                 ["T{},{}".format(origin[k], v)] +
                                                 ["X{},{},{}".format(i, j, v) for (i, j) in arc if (i, j, k) in arcs and i in ports] +
                                                 ["W{},{},{},{},{}".format(i, j, v, k, s) for (i, j) in arc for s in velocities if (i, j, k) in arcs] +
                                                 ["Y{},{}".format(v, k)],
                                              [1.0] +
                                              [-1.0] +
                                              [1.0] +
                                              [node_time[i] for (i, j) in arc if (i, j, k) in arcs and i in ports] +
                                              [sailing_time_2[i, j, k, s] for (i,j) in arc for s in velocities if (i, j, k) in arcs] +
                                              [(-1)*big_L[k]]] for k in vessels for v in voyages],
                                   senses=["G"]*vessel_count*voyage_count,
                                   rhs=[(-1)*big_L[k] for k in vessels for v in voyages])'''

    # Constraints 20
    for v in voyages:
        for k in vessels:
            ind = []
            val = []
            ind.append(name2idx["T{},{}".format(origin[k], v)])
            val.append(-1)
            ind.append(name2idx["Y{},{}".format(v, k)])
            val.append(vessel_available_at[k])
            problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                      val=val)],
                                           senses=["E"],
                                           rhs=[0])

    # Constraints 21
    for (i, j) in arc:
        for v in voyages:
            ind = []
            val = []
            ind.append(name2idx["T{},{}".format(i, v)])
            val.append(1)
            ind.append(name2idx["X{},{},{}".format(i, j, v)])
            val.append(node_time[i] if i in ports else 0)
            ind.append(name2idx["T{},{}".format(j, v)])
            val.append(-1)
            for s in velocities:
                for k in vessels:
                    if (i, j, v, k, s) in Wvar:
                        ind.append(name2idx["W{},{},{},{},{}".format(i, j, v, k, s)])
                        val.append(sailing_time_2[i, j, k, s])
            if i in ports:
                for c in contracts_U[i] + contracts_L[i]:
                    for p in products:
                        if (k, c, p) in Qvar:
                            ind.append(name2idx["Q{},{},{}".format(v, c, p)])
                            val.append(handling_times[p])
            problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                      val=val)],
                                           senses=["L"],
                                           rhs=[0])

    # Constraints 22
    for v in voyages:
        for c in transit_contracts:
            ind = []
            val = []
            ind.append(name2idx["T{},{}".format(load_port[c], v)])
            val.append(1)
            ind.append(name2idx["T{},{}".format(unload_port[c], v)])
            val.append(-1)
            ind.append(name2idx["D{},{}".format(v, c)])
            val.append((-1) * big_T[c])
            problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                      val=val)],
                                           senses=["G"],
                                           rhs=[(-1) * (big_T[c] + contract_transit_time[c])])

    # Constraints 23
    for j in ports:
        for k in vessels:
            for v in voyages:
                ind = []
                val = []
                ind.append(name2idx["T{},{}".format(j, v)])
                val.append(1)
                ind.append(name2idx["X{},{},{}".format(origin[k], j, v)])
                val.append(big_S[j, k])
                problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                          val=val)],
                                               senses=["L"],
                                               rhs=[(big_S[j, k] + end_of_planning_horizon)])

    # Constraints 24
    for k in vessels:
        for v in voyages:
            ind = []
            val = []
            ind.append(name2idx["WT{}".format(k)])
            val.append(1)
            ind.append(name2idx["T{},{}".format(destination[k], v)])
            val.append(-1)
            ind.append(name2idx["T{},{}".format(origin[k], v)])
            val.append(1)
            for (i, j) in arc:
                if (i, j, k) in arcs and i in ports:
                    ind.append(name2idx["X{},{},{}".format(i, j, v)])
                    val.append(node_time[i])
                for s in velocities:
                    if (i, j, k) in arcs:
                        ind.append(name2idx["W{},{},{},{},{}".format(i, j, v, k, s)])
                        val.append(sailing_time_2[i, j, k, s])
            ind.append(name2idx["Y{},{}".format(v, k)])
            val.append((-1) * big_L[k])
            problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                      val=val)],
                                           senses=["G"],
                                           rhs=[(-1) * big_L[k]])

    # -----------------------------------------------------------------------------------
    # Evenly Spread Constraints
    # -----------------------------------------------------------------------------------
    # Constraints 35
    '''problem.linear_constraints.add(lin_expr=[[["O{},{}".format(n, c) for n in range(int(minimum_pickups[c]), int(maximum_pickups[c]+1))] +
                                              ["S{}".format(c)] +
                                              ["Z{},{},{}".format(v, w, c)] +
                                              ["T{},{}".format(load_port[c], w)] +
                                              ["T{},{}".format(load_port[c], v)],
                                              [(end_of_planning_horizon)/n for n in range(int(minimum_pickups[c]), int(maximum_pickups[c]+1))] +
                                              [-1] +
                                              [big_E[c]] +
                                              [-1] +
                                              [1]]
                                             for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                                   senses=["L" for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                                   rhs=[big_E[c] for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)])

    # Constraints 36
    problem.linear_constraints.add(lin_expr=[[["O{},{}".format(n, c) for n in range(int(minimum_pickups[c]),int(maximum_pickups[c]+1))] +
                                              ["S{}".format(c)] +
                                              ["Z{},{},{}".format(v, w, c)] +
                                              ["T{},{}".format(load_port[c], w)] +
                                              ["T{},{}".format(load_port[c], v)],
                                              [(end_of_planning_horizon)/n for n in range(int(minimum_pickups[c]),int(maximum_pickups[c]+1))] +
                                              [1] +
                                              [(-1)*big_E[c]] +
                                              [-1] +
                                              [1]]
                                             for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                                   senses=["G" for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                                   rhs=[(-1)*big_E[c] for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)])

    # Constraints 37
    problem.linear_constraints.add(lin_expr=[[["S{}".format(c) for c in evenly_spread_contracts],
                                              [1.0]*(len(evenly_spread_contracts))]],
                                   senses=["L"],
                                   rhs=[service_level])

    # -----------------------------------------------------------------------------------
    # Valid Ineq
    # -----------------------------------------------------------------------------------
    # Constraints 25
    problem.linear_constraints.add(lin_expr=[[["WT{}".format(k) for k in vessels],
                                              [1]*vessel_count]],
                                   senses=["G"],
                                   rhs=[sum([2*handling_times[p]*contract_demand[c, p] for p in products for c in contracts if (c, p) in contract_demand])])'''


    '''problem.linear_constraints.add(lin_expr=[[["L{},{},{},{}".format(origin[k], j, v, p) for k in vessels for j in nodes if (origin[k], j, v, p) in Lvar],
                                              [1 for k in vessels for j in nodes if (origin[k], j, v, p) in Lvar]] for v in voyages for p in products],
                                   senses=["E"]*voyage_count*product_count,
                                   rhs=[0]*voyage_count*product_count)'''



    # Setting variables

    '''    for (i, j, v) in Xvar:
        ind = []
        val = []
        ind.append(name2idx["X{},{},{}".format(i, j, v)])
        val.append(-1)
        for k in vessels:
            for s in velocities:
                if (i, j, k) in arcs:
                    ind.append(name2idx["W{},{},{},{},{}".format(i, j, v, k, s)])
                    val.append(1)
        problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                  val=val)],
                                       senses=["E"],
                                       rhs=[0])'''
    # TODO: Trenger vi alle nullene?
    for key, value in DELTA.items():
        ind = []
        val = []
        ind.append(name2idx["D{},{}".format(key[0], key[1])])
        val.append(1)
        problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                  val=val)],
                                       senses=["E"],
                                       rhs=[value])

    '''problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)],
                                                  [1]] for (v, c) in Dvar],
                                       senses=["E" for d in Dvar],
                                       rhs=[(DELTA[v, c] if (v, c) in DELTA else 0) for (v, c) in Dvar])'''

    for key, value in Z.items():
        ind = []
        val = []
        ind.append(name2idx["Z{},{},{}".format(key[0], key[1], key[2])])
        val.append(1)
        problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                  val=val)],
                                       senses=["E"],
                                       rhs=[value])

    '''problem.linear_constraints.add(lin_expr=[[["Z{},{},{}".format(v, w, c)],
                                                  [1]] for (v, w, c) in Zvar],
                                       senses=["E" for z in z_variables],
                                       rhs=[Z[v, w, c] if (v, w, c) in Z else 0 for (v, w, c) in Zvar])'''

    for key, value in X.items():
        ind = []
        val = []
        ind.append(name2idx["X{},{},{}".format(key[0], key[1], key[2])])
        val.append(1)
        problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                  val=val)],
                                       senses=["E"],
                                       rhs=[value])


    '''problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(key[0], key[1], key[2])],
                                                  [1]] for key, value in X.items()],
                                       senses=["E" for key, value in X.items()],
                                       rhs=[value for key, value in X.items()])'''

    for key, value in phi.items():
        ind = []
        val = []
        ind.append(name2idx["O{},{}".format(key[0], key[1])])
        val.append(1)
        problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                  val=val)],
                                       senses=["E"],
                                       rhs=[value])

    '''problem.linear_constraints.add(lin_expr=[[["O{},{}".format(n, c)],
                                                  [1]] for (n, c) in Ovar],
                                   senses=["E" for (n, c) in Ovar],
                                   rhs=[phi[n, c] if (n, c) in phi else 0 for (n, c) in Ovar])'''

    for key, value in y.items():
        ind = []
        val = []
        ind.append(name2idx["Y{},{}".format(key[0], key[1])])
        val.append(1)
        problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                  val=val)],
                                       senses=["E"],
                                       rhs=[value])

    '''problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(key[0], key[1])],
                                                  [1]] for key, value in y.items()],
                                       senses=["E" for key, value in y.items()],
                                       rhs=[value for key, value in y.items()])'''

    # Writing the problem to text file
    #problem.write("voyage_model.lp")

    problem.set_log_stream(None)
    problem.set_error_stream(None)
    problem.set_warning_stream(None)
    problem.set_results_stream(None)

    print("CREATING MODEL: ", time.time()-t1)
    # Solve the problem
    start = problem.get_time()
    problem.solve()
    run_time = problem.get_time() - start
    print("RUN TIME: ", run_time)


    if problem.solution.get_status() == 103:
        return 103, [], None

    s_var = dict()

    for s in s_variables:
        s_var[s] = problem.solution.get_values(s)

    return problem.solution.get_status(), problem, s_var
