from input_for_voyage_model import *
import cplex

def fix_y_model(routes, cost_of_route, average_capacity_product, demand_in_port, route_visits_port, min_visits_products, contracts, visit_load_unload_port, min_sail_time_contract_route):
    problem = cplex.Cplex()
    problem.objective.set_sense(problem.objective.sense.minimize)
    binary = problem.variables.type.binary

    # OBJECTIVE FUNCTION
    problem.variables.add(obj=[(vessel_cost[k] * cost_of_route[r]) for k in vessels for r in routes],
                          names=["Y{},{}".format(r, k) for k in vessels for r, i in enumerate(routes)],
                          types=[binary] * len(vessels)*len(routes))

    name2idx = {n: j for j, n in enumerate(problem.variables.get_names())}

    # CONSTRAINTS
    problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(r, k) for r, i in enumerate(routes)],
                                              [1] * len(routes)] for k in vessels],
                                   senses=["E"]*len(vessels),
                                   rhs=[1]*len(vessels))

    problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(r, k) for k in vessels],
                                              [1] * len(vessels)] for r, i in enumerate(routes)],
                                   senses=["E"] * len(routes),
                                   rhs=[1] * len(routes))
    # TODO: CHECK
    '''problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(r, k) for k in vessels for r, _ in enumerate(routes)],
                                              [route_visits_port[i, r] * vessel_capacity[k, p] for k in vessels for r in routes]] for i in ports for p in products],
                                   senses=["G"] * len(ports) * len(products),
                                   rhs=[demand_in_port[i, p] for i in ports for p in products])'''

    problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(r, k) for k in vessels for r, _ in enumerate(routes)],
                                              [visit_load_unload_port[c, r] * vessel_capacity[k, p] for k in vessels for r in
                                               routes]] for c in contracts for p in products],
                                   senses=["G"] * len(contracts) * len(products),
                                   rhs=[contract_demand[c, p] if (c, p) in contract_demand else 0 for c in contracts for p in products])

    problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(r, k) for k in vessels for r, _ in enumerate(routes)],
                                              [vessel_capacity[k, p] * min_sail_time_contract_route[c, r] if (c, r)
                                                                   in min_sail_time_contract_route else 0 for k in vessels for
                                               r in
                                               routes]] for c in transit_contracts for p in products],
                                   senses=["G"] * len(transit_contracts) * len(products),
                                   rhs=[contract_demand[c, p] if (c, p) in contract_demand else 0 for c in transit_contracts for
                                        p in products])

    problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(r, k) for k in vessels for r, _ in enumerate(routes)],
                                              [(1 if (k, p) in vessel_capacity and vessel_capacity[k, p] > 0 else 0) *
                                                min_sail_time_contract_route[c, r] if (c, r) in min_sail_time_contract_route else 0
                                               for k in vessels for r in routes]] for c in transit_contracts for p in products],
                                   senses=["G"] * len(transit_contracts) * len(products),
                                   rhs=[minimum_pickups[c] if (c, p) in contract_demand else 0 for c in
                                        transit_contracts for p in products])

    problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(r, k) for k in vessels for r, _ in enumerate(routes)],
                                              [(visit_load_unload_port[c, r] * (1 if (k, p) in vessel_capacity and vessel_capacity[k, p] > 0 else 0))
                                               for k in vessels for r in routes]] for c in contracts for p in
                                             products],
                                   senses=["G"] * len(contracts) * len(products),
                                   rhs=[minimum_pickups[c] if (c, p) in contract_demand else 0 for c in
                                        contracts for p in
                                             products])

    '''problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(r, k) for k in vessels for r, _ in enumerate(routes)],
                                              [sum([vessel_capacity[k, q] if q >= p else 0 for q in products]) for k in vessels for
                                               r in
                                               routes]] for p in products],
                                   senses=["G"] * len(products),
                                   rhs=[sum([contract_demand[c, p] if (c, p) in contract_demand else 0 for c in contracts]) for p in products])'''

    #problem.write("y-model.lp")

    # STOP PRINTING
    problem.set_log_stream(None)
    problem.set_error_stream(None)
    problem.set_warning_stream(None)
    problem.set_results_stream(None)

    return problem