from input_for_voyage_model import *
import cplex
from config import *
import time


def route_model(cost_of_route, routes, minimum_visits_port, route_visits_port, ports, voyage_count, vessels,
                demand_in_port, average_capacity, demand_per_port, average_capacity_products,
                visit_load_unload_port, pickups, min_sail_time_contract_route):
    t0 = time.time()
    problem = cplex.Cplex()
    problem.objective.set_sense(problem.objective.sense.minimize)
    integer = problem.variables.type.integer

    # OBJECTIVE FUNCTION
    problem.variables.add(obj=[(cost_of_route[r]) for r in routes],
                          names=["X{}".format(r) for r in routes],
                          types=[integer] * len(routes),
                          ub=[voyage_count] * len(routes),
                          lb=[0] * len(routes))

    name2idx = {n: j for j, n in enumerate(problem.variables.get_names())}

    # CONSTRAINTS
    problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["X{}".format(r)] for r in routes],
                                                              val=[visit_load_unload_port[c, r] for r in routes]) for c in
                                             contracts],
                                   senses=["G"] * len(contracts),
                                   rhs=[pickups[c] for c in contracts])

    problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["X{}".format(r)] for r in routes],
                                             val=[1]*len(routes))],
                                   senses=["E"],
                                   rhs=[vessel_count])

    problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["X{}".format(r)] for r in routes],
                                                              val=[1 if (c, r)
                                                                   in min_sail_time_contract_route else 0 for r
                                                                   in routes]) for c in transit_contracts],
                                   senses=["G"] * len(transit_contracts),
                                   rhs=[pickups[c] for c in transit_contracts])

    '''problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["X{}".format(r)] for r in routes],
                                                              val=[sum([most_space[p][0] if r != 0 and (c, p) in
                                                                                                       contract_demand else 0 for c in contracts]) for i, r in enumerate(routes)]) for p in products],
                                   senses=["G"] * len(products),
                                   rhs=[sum([contract_demand[c, p] if (c,p) in contract_demand else 0 for c in contracts]) for p in products])'''



    #print("TIME: ", time.time()-t0)

    problem.set_log_stream(None)
    problem.set_error_stream(None)
    problem.set_warning_stream(None)
    problem.set_results_stream(None)

    problem.write("route_model.lp")

    return problem