from input_for_voyage_model import *
import cplex
from config import *
import time

def route_model_columns(routes, minimum_visits_port, route_visits_port, ports, voyage_count, vessels, demand_in_port):
    t0 = time.time()
    problem = cplex.Cplex()
    problem.objective.set_sense(problem.objective.sense.minimize)
    integer = problem.variables.type.integer

    # OBJECTIVE FUNCTION
    problem.variables.add(obj=[1 for r in routes],
                          names=["X{}".format(r) for r in routes],
                          types=[integer] * len(routes),
                          ub=[voyage_count] * len(routes),
                          lb=[0] * len(routes))

    name2idx = {n: j for j, n in enumerate(problem.variables.get_names())}

    # CONSTRAINTS
    problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["X{}".format(r)] for r in routes],
                                              val=[route_visits_port[i, r] for r in routes]) for i in ports],
                                   senses=["G"]*len(ports),
                                   rhs=[minimum_visits_port[i] for i in ports])

    problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["X{}".format(r)] for r in routes],
                                              val=[1]*len(routes))],
                                   senses=["L"],
                                   rhs=[vessel_count])

    print("TIME: ", time.time()-t0)

    problem.set_log_stream(None)
    problem.set_error_stream(None)
    problem.set_warning_stream(None)
    problem.set_results_stream(None)

    return problem