import cplex
from input_for_voyage_model import *
import time


def route_model_with_y(cost_of_route, routes, minimum_visits_port, route_visits_port, ports, voyage_count, vessels, demand_in_port, visit_load_unload_port, min_sail_time_contract_route):
    t0 = time.time()
    problem = cplex.Cplex()
    problem.objective.set_sense(problem.objective.sense.minimize)
    binary = problem.variables.type.binary

    problem.variables.add(obj=[(vessel_cost[k] * cost_of_route[r]) for r in routes for k in vessels],
                          names=["X{},{}".format(r, k) for r in routes for k in vessels],
                          types=[binary for r in routes for k in vessels])

    name2idx = {n: j for j, n in enumerate(problem.variables.get_names())}

    # Constraints
    '''problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["X{},{}".format(r, k)] for r in routes for k in vessels],
                                              val=[route_visits_port[i, r] for r in routes for k in vessels]) for i in ports],
                                   senses=["G"]*len(ports),
                                   rhs=[minimum_visits_port[i] for i in ports])'''

    problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=["X{},{}".format(r, k) for r in routes],
                                              val=[1]*len(routes)) for k in vessels],
                                   senses=["E"]*len(vessels),
                                   rhs=[1]*len(vessels))

    '''problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["X{},{}".format(r, k)] for r in routes],
                                              val=[1] * len(routes)) for k in vessels],
                                   senses=["L" for _ in vessels],
                                   rhs=[1 for _ in vessels])

    problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["X{},{}".format(r, k)] for r in routes for k in vessels],
                                              val=[route_visits_port[i, r]*vessel_capacity[k, p] for r in routes for k in vessels]) for i in ports for p in products],
                                   senses=["G"]*len(ports)*len(products),
                                   rhs=[demand_in_port[i, p] for i in ports for p in products])

    problem.linear_constraints.add(lin_expr=[[["X{},{}".format(r, k) for k in vessels for r in routes],
                                              [route_visits_port[i, r] * vessel_capacity[k, p] for k in vessels for r in
                                               routes]] for i in ports for p in products],
                                   senses=["G"] * len(ports) * len(products),
                                   rhs=[demand_in_port[i, p] for i in ports for p in products])'''

    problem.linear_constraints.add(lin_expr=[[["X{},{}".format(r, k) for k in vessels for r in routes],
                                              [visit_load_unload_port[c, r] * vessel_capacity[k, p] for k in vessels for
                                               r in
                                               routes]] for c in contracts for p in products],
                                   senses=["G"] * len(contracts) * len(products),
                                   rhs=[contract_demand[c, p] if (c, p) in contract_demand else 0 for c in contracts for
                                        p in products])

    problem.linear_constraints.add(lin_expr=[[["X{},{}".format(r, k) for k in vessels for r in routes],
                                              [visit_load_unload_port[c, r] for k in vessels for
                                               r in
                                               routes]] for c in contracts],
                                   senses=["G"] * len(contracts),
                                   rhs=[minimum_pickups[c] for c in contracts])

    problem.linear_constraints.add(lin_expr=[[["X{},{}".format(r, k) for k in vessels for r in routes],
                                              [vessel_capacity[k, p] * min_sail_time_contract_route[c, r] if (c, r)
                                                                                                             in min_sail_time_contract_route else 0
                                               for k in vessels for
                                               r in
                                               routes]] for c in transit_contracts for p in products],
                                   senses=["G"] * len(transit_contracts) * len(products),
                                   rhs=[contract_demand[c, p] if (c, p) in contract_demand else 0 for c in
                                        transit_contracts for
                                        p in products])

    problem.linear_constraints.add(lin_expr=[[["X{},{}".format(r, k) for k in vessels for r in routes],
                                              [(1 if (k, p) in vessel_capacity and vessel_capacity[k, p] > 0 else 0) *
                                               min_sail_time_contract_route[c, r] if (c,
                                                                                      r) in min_sail_time_contract_route else 0
                                               for k in vessels for r in routes]] for c in transit_contracts for p in
                                             products],
                                   senses=["G"] * len(transit_contracts) * len(products),
                                   rhs=[minimum_pickups[c] if (c, p) in contract_demand else 0 for c in
                                        transit_contracts for p in products])

    problem.linear_constraints.add(lin_expr=[[["X{},{}".format(r, k) for k in vessels for r in routes],
                                              [sum([vessel_capacity[k, q] if q >= p else 0 for q in products]) for k in
                                               vessels for
                                               r in
                                               routes]] for p in products],
                                   senses=["G"] * len(products),
                                   rhs=[sum(
                                       [contract_demand[c, p] if (c, p) in contract_demand else 0 for c in contracts])
                                        for p in products])
    t1 = time.time()
    print("TIME CREATE ROUTE MODEL: ", t1-t0)

    problem.set_log_stream(None)
    problem.set_error_stream(None)
    problem.set_warning_stream(None)
    problem.set_results_stream(None)


    '''problem.write("route.lp")

    problem.solve()
    sol = {}
    for r in routes:
        for k in vessels:
            if problem.solution.get_values("X{},{}".format(r, k)) == 1.0:
                sol[r, k] = 1

    print("SOLUTION: ", sol)'''

    #return sol, problem.solution.get_objective_value()
    return problem

