from input_for_voyage_model import *
import cplex
from config import *
import time
from Visual import Visual


problem = cplex.Cplex()
#problem.set_problem_name("Voyage model")
problem.objective.set_sense(problem.objective.sense.minimize)
binary = problem.variables.type.binary

# -----------------------------------------------------------------------------------
# Parameters
# -----------------------------------------------------------------------------------
service_level = 30

# -----------------------------------------------------------------------------------
# Variables
# -----------------------------------------------------------------------------------
x_variables = ["X{},{},{}".format(i, j, v) for (i, j, v) in Xvar]
w_variables = ["W{},{},{},{},{}".format(i, j, v, k, s) for (i, j, v, k, s) in Wvar]
l_variables = ["L{},{},{},{}".format(i, j, v, p) for (i, j, v, p) in Lvar]
y_variables = ["Y{},{}".format(v, k) for (v, k) in Yvar]
q_variables = ["Q{},{},{}".format(v, c, p) for (v, c, p) in Qvar]
d_variables = ["D{},{}".format(v, c) for (v, c) in Dvar]
t_variables = ["T{},{}".format(i, v) for (i, v) in Tvar]
z_variables = ["Z{},{},{}".format(v, w, c) for (v, w, c) in Zvar]
s_variables = ["S{}".format(c) for c in Svar]
wt_variables = ["WT{}".format(k) for k in WTvar]

o_variables = ["O{},{}".format(n, c) for (n, c) in Ovar]

'''problem.variables.add(names=x_variables,
                      types=[binary for i in range(len(x_variables))])'''
problem.variables.add(names=l_variables)
problem.variables.add(names=y_variables)
problem.variables.add(names=y_variables,
                      types=[binary for i in range(len(Yvar))])
problem.variables.add(names=q_variables)
problem.variables.add(names=t_variables)
'''problem.variables.add(names=w_variables,
                      lb=[0]*len(Wvar),
                      ub=[1]*len(Wvar))'''
problem.variables.add(names=d_variables,
                      types=[binary for i in range(len(Dvar))])

problem.variables.add(names=z_variables,
                      types=[binary for i in range(len(Zvar))])
problem.variables.add(names=s_variables,
                      lb=[0 for c in evenly_spread_contracts])
'''problem.variables.add(names=wt_variables)'''
problem.variables.add(names=o_variables,
                      types=[binary for i in range(len(Ovar))])

# -----------------------------------------------------------------------------------
# Objective Function
# -----------------------------------------------------------------------------------
problem.variables.add(obj=[leg_sailing_cost[i, j, k, s] + charter_cost[k] * (node_time[i] if i in node_time else 0) for (i, j, v,  k, s) in Wvar],
                      names=["W{},{},{},{},{}".format(i, j, v, k, s) for (i, j, v, k, s) in Wvar])
problem.variables.add(obj=[port_cost[i] if i in ports else 0 for i in nodes for j in nodes for v in voyages if (i, j, v) in Xvar],
                      names=["X{},{},{}".format(i, j, v) for i in nodes for j in nodes for v in voyages if (i, j, v) in Xvar],
                      types=[binary for i in nodes for j in nodes for v in voyages if (i, j, v) in Xvar])
problem.variables.add(obj=[charter_cost[k] for k in vessels],
                      names=["WT{}".format(k) for k in vessels])


# -----------------------------------------------------------------------------------
# Network Constraints
# -----------------------------------------------------------------------------------
t0 = time.time()
# Constraints 2
problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(origin[k], j, v)
                                            for k in vessels for j in ports + [destination[k]]],
                                          [1]*(port_count+1)*vessel_count] for v in voyages],
                               senses=["E"]*voyage_count,
                               rhs=[1]*voyage_count)


# Constraints 3
problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(i, j, v) for i in nodes if i < j] +
                                          ["X{},{},{}".format(j, i, v) for i in nodes if i > j],
                                          [1.0 for i in nodes if i < j] +
                                          [(-1.0) for i in nodes if i > j]] for j in ports for v in voyages],
                               senses=["E"]*port_count*voyage_count,
                               rhs=[0]*port_count*voyage_count)

# Constraints 4
problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(i, destination[k], v)
                                           for k in vessels for i in ports + [origin[k]]],
                                          [1]*(port_count+1)*vessel_count] for v in voyages],
                               senses=["E"]*voyage_count,
                               rhs=[1]*voyage_count)

# -----------------------------------------------------------------------------------
#  Vessel Constraints
# -----------------------------------------------------------------------------------
# Constraints 5
problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(i, j, v)] +
                                          ["W{},{},{},{},{}".format(i, j, v, k, s)
                                           for s in velocities for k in vessels if (i, j, k) in arcs],
                                          [-1.0] +
                                          [1.0 for s in velocities for k in vessels if (i, j, k) in arcs]] for (i, j, v) in Xvar],
                               senses=["E"]*len(Xvar),
                               rhs=[0]*len(Xvar))

# Constraints 8
problem.linear_constraints.add(lin_expr=[[["W{},{},{},{},{}".format(i, j, v, k, s)
                                            for s in velocities if (i,j,k) in arcs] +
                                             ["Y{},{}".format(v, k)],
                                          [1 for s in velocities if (i, j, k) in arcs] +
                                          [-1]] for (i, j, v) in Xvar for k in vessels],
                               senses=["L"]*vessel_count*len(Xvar),
                               rhs=[0]*vessel_count*len(Xvar))

# Constraints 9
problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(v, k)
                                            for v in voyages],
                                          [1]*voyage_count] for k in vessels],
                               senses=["E"]*vessel_count,
                               rhs=[1]*vessel_count)

# Constraints 10
problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(v, k)
                                            for k in vessels],
                                          [1]*vessel_count] for v in voyages],
                               senses=["E"]*voyage_count,
                               rhs=[1]*voyage_count)

# -----------------------------------------------------------------------------------
# Capacity Constraints
# -----------------------------------------------------------------------------------
# Constraints 11
problem.linear_constraints.add(lin_expr=[[["L{},{},{},{}".format(i, j, v, p)] +
                                             ["Y{},{}".format(v, k) for k in vessels] +
                                             ["L{},{},{},{}".format(i, j, v, q) for q in products_S[p]],
                                          [-1] +
                                          [vessel_capacity[k, p] for k in vessels] +
                                          [-1]*len(products_S[p])] for p in products for (i, j) in arc for v in voyages],
                               senses=["G"]*product_count*len(arc)*voyage_count,
                               rhs=[0]*product_count*len(arc)*voyage_count,
                               names=["Constraints11_{}".format(i) for i in range(1, product_count*len(arc)*voyage_count+1)])

# Constraints 12
problem.linear_constraints.add(lin_expr=[[["L{},{},{},{}".format(i, j, v, p)] +
                                            ["X{},{},{}".format(i, j, v)],
                                          [1] +
                                          [(-1)*max_capacity[p]]] for i in ports for j in nodes for v in voyages for p in products if (i, j, v) in Xvar],
                               senses=["L" for i in ports for j in nodes for v in voyages for p in products  if (i, j, v) in Xvar],
                               rhs=[0 for i in ports for j in nodes for v in voyages for p in products  if (i, j, v) in Xvar],
                               names=["HER_{}".format(i) for i in ports for j in nodes for v in voyages for p in products  if (i, j, v) in Xvar])

# Constraints 13
problem.linear_constraints.add(lin_expr=[[["L{},{},{},{}".format(j, i, v, p) for j in nodes if (j, i, v, p) in Lvar] +
                                            ["Q{},{},{}".format(v, c, p) for c in (contracts_L[i] if i in ports else []) if (v, c, p) in Qvar] +
                                             ["Q{},{},{}".format(v, c, p) for c in (contracts_U[i] if i in ports else []) if (v, c, p) in Qvar] +
                                             ["L{},{},{},{}".format(i, j, v, p) for j in nodes if (i, j, v, p) in Lvar],
                                          [1 for j in nodes if (j, i, v, p) in Lvar] +
                                          [1 for c in (contracts_L[i] if i in ports else []) if (v, c, p) in Qvar] +
                                          [-1 for c in (contracts_U[i] if i in ports else []) if (v, c, p) in Qvar] +
                                          [-1 for j in nodes if (i, j, v, p) in Lvar]] for i in nodes for v in vessels for p in products],
                               senses=["E"]*product_count*last_node*vessel_count,
                               rhs=[0]*product_count*last_node*vessel_count)

# -----------------------------------------------------------------------------------
# Pickup Constraints
# -----------------------------------------------------------------------------------
# Constraints 15
problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)
                                            for v in voyages],
                                          [1]*voyage_count] for c in evenly_spread_contracts],
                               senses=["L"]*len(evenly_spread_contracts),
                               rhs=[maximum_pickups[c] for c in evenly_spread_contracts])


problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)
                                            for v in voyages],
                                          [1]*voyage_count] for c in evenly_spread_contracts],
                               senses=["G"]*len(evenly_spread_contracts),
                               rhs=[minimum_pickups[c] for c in evenly_spread_contracts])

# Constraints 16
problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)] +
                                          ["X{},{},{}".format(i, load_port[c], v) for i in nodes if (i, load_port[c]) in arc],
                                          [1] +
                                          [-1 for i in nodes if (i, load_port[c]) in arc]] for c in contracts for v in voyages],
                               senses=["L"]*contract_count*voyage_count,
                               rhs=[0]*contract_count*voyage_count)


# Constraints 17
problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)] +
                                            ["X{},{},{}".format(i, unload_port[c], v) for i in nodes if (i, unload_port[c]) in arc],
                                          [1] +
                                          [-1 for i in nodes if (i, unload_port[c]) in arc]] for c in contracts for v in voyages],
                               senses=["L"]*contract_count*voyage_count,
                               rhs=[0]*contract_count*voyage_count)


# -----------------------------------------------------------------------------------
# Loading Constraints
# -----------------------------------------------------------------------------------
# Constraints 18
problem.linear_constraints.add(lin_expr=[[["Q{},{},{}".format(v, c, p)] +
                                             ["D{},{}".format(v, c)],
                                          [1] +
                                          [(-1)*minimum_quantity[c, p]]] for c in contracts for v in voyages for p in products if (c,p) in maximum_quantity],
                               senses=["G" for c in contracts for v in voyages for p in products if (c,p) in maximum_quantity],
                               rhs=[0 for c in contracts for v in voyages for p in products if (c,p) in maximum_quantity])

problem.linear_constraints.add(lin_expr=[[["Q{},{},{}".format(v, c, p)] +
                                             ["D{},{}".format(v, c)],
                                          [1] +
                                          [(-1)*maximum_quantity[c,p]]] for c in contracts for v in voyages for p in products if (c, p) in maximum_quantity],
                               senses=["L" for c in contracts for v in voyages for p in products if (c,p) in maximum_quantity],
                               rhs=[0 for c in contracts for v in voyages for p in products if (c,p) in maximum_quantity])

problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)] +
                                          ["Q{},{},{}".format(v, c, p)],
                                          [(-1)*maximum_quantity[c, p]] +
                                          [1]] for (v, c, p) in Qvar if (c, p) in maximum_quantity],
                               senses=["L" for (v, c, p) in Qvar if (c, p) in maximum_quantity],
                               rhs=[0 for (v, c, p) in Qvar if (c, p) in maximum_quantity])

problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)] +
                                          ["Q{},{},{}".format(v, c, p)],
                                          [(-1)*minimum_quantity[c, p]] +
                                          [1]] for (v, c, p) in Qvar if (c, p) in minimum_quantity],
                               senses=["G" for (v, c, p) in Qvar if (c, p) in minimum_quantity],
                               rhs=[0 for (v, c, p) in Qvar if (c, p) in minimum_quantity])



# Constraints 19 NB: E i xpress
problem.linear_constraints.add(lin_expr=[[["Q{},{},{}".format(v, c, p) for v in voyages],
                                          [1 for v in voyages]] for c in contracts for p in products if (c, p) in contract_demand and contract_demand[c, p] > 0],
                               senses=["E" for c in contracts for p in products if (c, p) in contract_demand and contract_demand[c, p] > 0],
                               rhs=[contract_demand[c, p] for c in contracts for p in products if (c, p) in contract_demand and contract_demand[c, p] > 0])


'''problem.linear_constraints.add(lin_expr=[[["Q{},{},{}".format(k, c, p) for k in range(1, vessel_count+1)],
                                          [1.0]*vessel_count] for (c, p) in cp if (c, p) in contract_demand and contract_demand[c, p] > 0],
                               senses=["E"]*len([x for (c, p) in cp if (c, p) in contract_demand and contract_demand[c, p] > 0]),
                               rhs=[contract_demand[c, p] for (c, p) in cp if (c, p) in contract_demand and contract_demand[c, p] > 0])'''

# -----------------------------------------------------------------------------------
# Time Constraints
# -----------------------------------------------------------------------------------
# Constraints 20
problem.linear_constraints.add(lin_expr=[[["T{},{}".format(origin[k], v)] +
                                             ["Y{},{}".format(v, k)],
                                          [-1] +
                                          [vessel_available_at[k]]] for v in voyages for k in vessels],
                               senses=["E"]*voyage_count*vessel_count,
                               rhs=[0]*voyage_count*vessel_count)

# Constraints 21
problem.linear_constraints.add(lin_expr=[[["T{},{}".format(i, v)] +
                                          ["W{},{},{},{},{}".format(i, j, v, k, s) for s in velocities for k in vessels if (i, j, v, k, s) in Wvar] +
                                          ["X{},{},{}".format(i, j, v)] +
                                          ["Q{},{},{}".format(v, c, p) for c in (contracts_U[i]+contracts_L[i] if i in ports else []) for p in products if (k, c, p) in Qvar] +
                                          ["T{},{}".format(j, v)],
                                          [1.0] +
                                          [sailing_time_2[i, j, k, s] for s in velocities for k in vessels if (i, j, v, k, s) in Wvar] +
                                          [node_time[i] if i in ports else 0] +
                                          [handling_times[p] for c in (contracts_U[i]+contracts_L[i] if i in ports else [])for p in products if (k, c, p) in Qvar] +
                                          [-1.0]] for (i, j) in arc for v in voyages],
                               senses=["L"]*len(arc)*voyage_count,
                               rhs=[0]*len(arc)*voyage_count)

# Constraints 22
problem.linear_constraints.add(lin_expr=[[["T{},{}".format(load_port[c], v)] +
                                          ["T{},{}".format(unload_port[c], v)] +
                                          ["D{},{}".format(v, c)],
                                          [1.0] +
                                          [-1.0] +
                                          [(-1)*big_T[c]]] for v in voyages for c in transit_contracts],
                               senses=["G"]*voyage_count*len(transit_contracts),
                               rhs=[(-1)*(big_T[c] + contract_transit_time[c]) for v in voyages for c in transit_contracts])

# Constraints 23
problem.linear_constraints.add(lin_expr=[[["T{},{}".format(j, v)] +
                                          ["X{},{},{}".format(origin[k], j, v)],
                                          [1.0] +
                                          [big_S[j, k]]] for j in ports for k in vessels for v in voyages],
                               senses=["L"]*port_count*vessel_count*voyage_count,
                               rhs=[(big_S[j, k] + end_of_planning_horizon) for j in ports for k in vessels for v in voyages])

# Constraints 24
problem.linear_constraints.add(lin_expr=[[["WT{}".format(k)] +
                                          ["T{},{}".format(destination[k], v)] +
                                             ["T{},{}".format(origin[k], v)] +
                                             ["X{},{},{}".format(i, j, v) for (i, j) in arc if (i, j, k) in arcs and i in ports] +
                                             ["W{},{},{},{},{}".format(i, j, v, k, s) for (i, j) in arc for s in velocities if (i, j, k) in arcs] +
                                             ["Y{},{}".format(v, k)],
                                          [1.0] +
                                          [-1.0] +
                                          [1.0] +
                                          [node_time[i] for (i, j) in arc if (i, j, k) in arcs and i in ports] +
                                          [sailing_time_2[i, j, k, s] for (i,j) in arc for s in velocities if (i, j, k) in arcs] +
                                          [(-1)*big_L[k]]] for k in vessels for v in voyages],
                               senses=["G"]*vessel_count*voyage_count,
                               rhs=[(-1)*big_L[k] for k in vessels for v in voyages],
                               names=["mjoff{}".format(i) for i in range(vessel_count*voyage_count)])

# -----------------------------------------------------------------------------------
# Evenly Spread Constraints
# -----------------------------------------------------------------------------------
# Constraints 30
problem.linear_constraints.add(lin_expr=[[["Z{},{},{}".format(v, w, c) for v in voyages for w in voyages if (v, w, c) in Zvar] +
                                          ["D{},{}".format(v, c) for v in voyages],
                                          [1 for w in voyages for v in voyages if (v, w, c) in Zvar] +
                                          [-1] * voyage_count] for c in evenly_spread_contracts],
                               senses=["G"]*len(evenly_spread_contracts),
                               rhs=[-1.0]*len(evenly_spread_contracts))

# Constraints 31
problem.linear_constraints.add(lin_expr=[[["Z{},{},{}".format(v, w, c) for w in range((v+1), voyage_count + 1)] +
                                          ["D{},{}".format(v, c)],
                                          [1 for w in range((v+1), voyage_count + 1)] +
                                          [-1]] for c in evenly_spread_contracts for v in voyages],
                               senses=["L"]*len(evenly_spread_contracts)*voyage_count,
                               rhs=[0.0]*len(evenly_spread_contracts)*voyage_count)

# Constraints 32
problem.linear_constraints.add(lin_expr=[[["Z{},{},{}".format(w, v, c) for w in range(1, v)] +
                                          ["D{},{}".format(v, c)],
                                          [1  for w in range(1, v)] +
                                          [-1]] for c in evenly_spread_contracts for v in voyages],
                               senses=["L"]*len(evenly_spread_contracts)*voyage_count,
                               rhs=[0.0]*len(evenly_spread_contracts)*voyage_count)


# Constraints 33
problem.linear_constraints.add(lin_expr=[[["O{},{}".format(n, c) for n in range(int(minimum_pickups[c]), int(maximum_pickups[c]+1))] +
                                          ["D{},{}".format(v, c) for v in voyages],
                                          [n for n in range(int(minimum_pickups[c]), int(maximum_pickups[c]+1))] +
                                          [-1] * voyage_count] for c in evenly_spread_contracts],
                               senses=["E"]*len(evenly_spread_contracts),
                               rhs=[0.0]*len(evenly_spread_contracts))

# Constraints 34
problem.linear_constraints.add(lin_expr=[[["O{},{}".format(n, c) for n in range(int(minimum_pickups[c]),int(maximum_pickups[c]+1))],
                                          [1.0]*int(maximum_pickups[c]-minimum_pickups[c]+1)] for c in evenly_spread_contracts],
                               senses=["E"]*len(evenly_spread_contracts),
                               rhs=[1.0]*len(evenly_spread_contracts))

# Constraints 35
problem.linear_constraints.add(lin_expr=[[["O{},{}".format(n, c) for n in range(int(minimum_pickups[c]),int(maximum_pickups[c]+1))] +
                                          ["S{}".format(c)] +
                                          ["Z{},{},{}".format(v, w, c)] +
                                          ["T{},{}".format(load_port[c], w)] +
                                          ["T{},{}".format(load_port[c], v)],
                                          [(end_of_planning_horizon)/n for n in range(int(minimum_pickups[c]),int(maximum_pickups[c]+1))] +
                                          [-1] +
                                          [big_E[c]] +
                                          [-1] +
                                          [1]]
                                         for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                               senses=["L" for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                               rhs=[big_E[c] for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)])

# Constraints 36
problem.linear_constraints.add(lin_expr=[[["O{},{}".format(n, c) for n in range(int(minimum_pickups[c]), int(maximum_pickups[c]+1))] +
                                          ["S{}".format(c)] +
                                          ["Z{},{},{}".format(v, w, c)] +
                                          ["T{},{}".format(load_port[c], w)] +
                                          ["T{},{}".format(load_port[c], v)],
                                          [(end_of_planning_horizon)/n for n in range(int(minimum_pickups[c]), int(maximum_pickups[c]+1))] +
                                          [1] +
                                          [(-1)*big_E[c]] +
                                          [-1] +
                                          [1]]
                                         for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                               senses=["G" for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                               rhs=[(-1)*big_E[c] for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)])

# Constraints 37
problem.linear_constraints.add(lin_expr=[[["S{}".format(c) for c in evenly_spread_contracts],
                                          [1.0]*(len(evenly_spread_contracts))]],
                               senses=["L"],
                               rhs=[service_level])

# -----------------------------------------------------------------------------------
# Valid Ineq
# -----------------------------------------------------------------------------------
# Constraints 6
problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(origin[k], j, v)
                                            for v in voyages for j in ports+[destination[k]]],
                                          [1]*voyage_count*(port_count+1)] for k in vessels],
                               senses=["E"]*vessel_count,
                               rhs=[1]*vessel_count)

# Constraints 7
problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(i, destination[k], v)
                                            for v in voyages for i in ports+[origin[k]]],
                                          [1]*voyage_count*(port_count+1)] for k in vessels],
                               senses=["E"]*vessel_count,
                               rhs=[1]*vessel_count)

# Constraints 26
problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(origin[k], j, (v+1))
                                            for j in ports for k in vessels] +
                                             ["X{},{},{}".format(origin[k], j, v) for j in ports for k in vessels],
                                          [1]*vessel_count*port_count +
                                          [-1]*vessel_count*port_count] for v in voyages if not voyage_count],
                               senses=["L"]*(voyage_count-1),
                               rhs=[0]*(voyage_count-1))

# Constraints 25
problem.linear_constraints.add(lin_expr=[[["WT{}".format(k) for k in vessels],
                                          [1]*vessel_count]],
                               senses=["G"],
                               rhs=[sum([2*handling_times[p]*contract_demand[c, p] for p in products for c in contracts if (c, p) in contract_demand])])



problem.linear_constraints.add(lin_expr=[[["L{},{},{},{}".format(origin[k], j, v, p) for k in vessels for j in nodes if (origin[k], j, v, p) in Lvar],
                                          [1 for k in vessels for j in nodes if (origin[k], j, v, p) in Lvar]] for v in voyages for p in products],
                               senses=["E"]*voyage_count*product_count,
                               rhs=[0]*voyage_count*product_count)
t1 = time.time()
print("TIME: ", t1-t0)
# Writing the problem to text file
#problem.write("voyage_model.lp")

# Solve the problem
start = problem.get_time()
problem.solve()
run_time = problem.get_time() - start


print(" ------------------------------------------------------ ")
print(" ------------------ VOYAGE MODEL ---------------------- ")
print(" ------------------------------------------------------ ")

print("----------------------------------")
print("Status: ", problem.solution.get_status_string(problem.solution.get_status()))
print("Objective: ", "{:,}".format(round(problem.solution.get_objective_value())))
print("Time: ", run_time)
print("----------------------------------")

d_var = {}
for v in voyages:
    for c in contracts:
        d_var[v, c] = problem.solution.get_values("D{},{}".format(v, c))
'''
for x in x_variables:
    if problem.solution.get_values(x) != 0:
        print(x, problem.solution.get_values(x))

for d in d_variables:
    if problem.solution.get_values(d) != 0:
        print(d, problem.solution.get_values(d))

for y in y_variables:
    if problem.solution.get_values(y) != 0:
        print(y, problem.solution.get_values(y))

for z in z_variables:
    if problem.solution.get_values(z) != 0:
        print(z, problem.solution.get_values(z))'''

x_var = {}
for v in voyages:
    for i in nodes:
        for j in nodes:
            if (i, j, v) in Xvar:
                x_var[i, j, v] = problem.solution.get_values("X{},{},{}".format(i, j, v))

t_var = {}
for v in voyages:
    for i in ports:
        t_var[i, v] = problem.solution.get_values("T{},{}".format(i, v))

y_var = {}
for v in voyages:
    for k in vessels:
        y_var[v, k] = problem.solution.get_values("Y{},{}".format(v, k))

l_var = {}
for v in voyages:
    for p in products:
        for i in ports:
            for j in ports:
                if i < j:
                    l_var[i, j, v, p] = problem.solution.get_values("L{},{},{},{}".format(i, j, v, p))

for s in s_variables:
    print(s, problem.solution.get_values(s))

#vis = Visual(x_var, t_var, l_var, y_var)


with open('results/results.txt', 'a') as f:
    f.write(FILENAME)
    f.write("\t\tVoyage model\t")
    f.write(str(run_time) + "\t\t")
    f.write(str(round(problem.solution.get_objective_value())))
    f.write("\t")
    f.write(problem.solution.get_status_string(problem.solution.get_status()))
    f.write("\n")

with open('results/voyage_model/delta.txt', 'a') as f:
    f.write(FILENAME + "\t\t")
    f.write(str(run_time) + "\t\t")
    f.write(str(round(problem.solution.get_objective_value())) + "\t\t")
    f.write("delta = {")
    index = 0
    for key, d in d_var.items():
        index += 1
        f.write(str(key) + ": ")
        f.write(str(d))
        if index != len(d_var):
            f.write(", ")
    f.write("}")
    f.write("\n")

with open('results/voyage_model/y-var.txt', 'a') as f:
    f.write(FILENAME + "\t\t")
    f.write(str(run_time) + "\t\t")
    f.write(str(round(problem.solution.get_objective_value())) + "\t\t")
    f.write("y = {")
    index = 0
    for key, d in y_var.items():
        index += 1
        f.write(str(key) + ": ")
        f.write(str(d))
        if index != len(d_var):
            f.write(", ")
    f.write("}")
    f.write("\n")
