from input_for_voyage_model import *
import cplex
import time
from config import service_level, penalty

def model_delta_fixed():
    t1 = time.time()
    problem = cplex.Cplex()
    problem.objective.set_sense(problem.objective.sense.minimize)
    binary = problem.variables.type.binary

    # -----------------------------------------------------------------------------------
    # Parameters
    # -----------------------------------------------------------------------------------

    # set gap
    #problem.parameters.mip.tolerances.mipgap.set(0.01)

    # -----------------------------------------------------------------------------------
    # Variables
    # -----------------------------------------------------------------------------------
    x_variables = ["X{},{},{}".format(i, j, v) for (i, j, v) in Xvar]
    #w_variables = ["W{},{},{},{},{}".format(i, j, v, k, s) for (i, j, v, k, s) in Wvar]
    l_variables = ["L{},{},{},{}".format(i, j, v, p) for (i, j, v, p) in Lvar]
    y_variables = ["Y{},{}".format(v, k) for (v, k) in Yvar]
    q_variables = ["Q{},{},{}".format(v, c, p) for (v, c, p) in Qvar]
    d_variables = ["D{},{}".format(v, c) for (v, c) in Dvar]
    t_variables = ["T{},{}".format(i, v) for (i, v) in Tvar]
    z_variables = ["Z{},{},{}".format(v, w, c) for (v, w, c) in Zvar]
    #s_variables = ["S{}".format(c) for c in Svar]
    #wt_variables = ["WT{}".format(k) for k in WTvar]

    o_variables = ["O{},{}".format(n, c) for (n, c) in Ovar]

    #problem.variables.add(names=x_variables)
     #                     #types=[binary for i in range(len(x_variables))])
    problem.variables.add(names=l_variables)
    problem.variables.add(names=y_variables)
                          #types=[binary for i in range(len(Yvar))])
    problem.variables.add(names=q_variables)
    problem.variables.add(names=t_variables)
    '''problem.variables.add(names=w_variables,
                          lb=[0]*len(Wvar),
                          ub=[1]*len(Wvar))'''
    problem.variables.add(names=d_variables)
                          #types=[binary for i in range(len(Dvar))])

    problem.variables.add(names=z_variables)
                          #types=[binary for i in range(len(Zvar))])
    #problem.variables.add(names=s_variables,
     #                     lb=[0 for c in evenly_spread_contracts])
    '''problem.variables.add(names=wt_variables)'''
    problem.variables.add(names=o_variables)
                          #types=[binary for i in range(len(Ovar))])

    # -----------------------------------------------------------------------------------
    # Objective Function
    # -----------------------------------------------------------------------------------
    problem.variables.add(obj=[leg_sailing_cost[i, j, k, s] + charter_cost[k] * (node_time[i] if i in node_time else 0) for (i, j, v,  k, s) in Wvar],
                          names=["W{},{},{},{},{}".format(i, j, v, k, s) for (i, j, v, k, s) in Wvar],
                          lb=[0] * len(Wvar),
                          ub=[1] * len(Wvar)
                          )
    problem.variables.add(obj=[port_cost[i] if i in ports else 0 for i in nodes for j in nodes for v in voyages if (i, j, v) in Xvar],
                          names=["X{},{},{}".format(i, j, v) for i in nodes for j in nodes for v in voyages if (i, j, v) in Xvar])
                          #types=[binary for i in ports for j in nodes for v in voyages if (i, j, v) in Xvar])
    problem.variables.add(obj=[charter_cost[k] for k in vessels],
                          names=["WT{}".format(k) for k in vessels])
    problem.variables.add(obj=[penalty for c in Svar],
                          names=["S{}".format(c) for c in Svar])

    name2idx = {n: j for j, n in enumerate(problem.variables.get_names())}

    problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(origin[k], j, v)
                                               for k in vessels for j in ports + [destination[k]]],
                                              [1] * (port_count + 1) * vessel_count] for v in voyages],
                                   senses=["E"] * voyage_count,
                                   rhs=[1] * voyage_count)

    # Constraints 3
    problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(i, j, v) for i in nodes if i < j] +
                                              ["X{},{},{}".format(j, i, v) for i in nodes if i > j],
                                              [1.0 for i in nodes if i < j] +
                                              [(-1.0) for i in nodes if i > j]] for j in ports for v in voyages],
                                   senses=["E"] * port_count * voyage_count,
                                   rhs=[0] * port_count * voyage_count)

    # Constraints 4
    problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(i, destination[k], v)
                                               for k in vessels for i in ports + [origin[k]]],
                                              [1] * (port_count + 1) * vessel_count] for v in voyages],
                                   senses=["E"] * voyage_count,
                                   rhs=[1] * voyage_count)

    # -----------------------------------------------------------------------------------
    #  Vessel Constraints
    # -----------------------------------------------------------------------------------
    # Constraints 5
    for (i, j, v) in Xvar:
        ind = []
        val = []
        ind.append(name2idx["X{},{},{}".format(i, j, v)])
        val.append(-1)
        for k in vessels:
            for s in velocities:
                if (i, j, k) in arcs:
                    ind.append(name2idx["W{},{},{},{},{}".format(i, j, v, k, s)])
                    val.append(1)
        problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                  val=val)],
                                       senses=["E"],
                                       rhs=[0])

    # Constraints 8
    for (i, j, v) in Xvar:
        for k in vessels:
            ind = []
            val = []
            if (i, j, k) in arcs:
                for s in velocities:
                    ind.append(name2idx["W{},{},{},{},{}".format(i, j, v, k, s)])
                    val.append(1)
            ind.append(name2idx["Y{},{}".format(v, k)])
            val.append(-1)
            problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                      val=val)],
                                           senses=["L"],
                                           rhs=[0])

    # Constraints 9
    problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(v, k)
                                               for v in voyages],
                                              [1] * voyage_count] for k in vessels],
                                   senses=["E"] * vessel_count,
                                   rhs=[1] * vessel_count)

    # Constraints 10
    problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(v, k)
                                               for k in vessels],
                                              [1] * vessel_count] for v in voyages],
                                   senses=["E"] * voyage_count,
                                   rhs=[1] * voyage_count)

    # -----------------------------------------------------------------------------------
    # Capacity Constraints
    # -----------------------------------------------------------------------------------
    # Constraints 11
    for p in products:
        for (i, j) in arc:
            for v in voyages:
                ind = []
                val = []
                ind.append(name2idx["L{},{},{},{}".format(i, j, v, p)])
                val.append(-1)
                for k in vessels:
                    ind.append(name2idx["Y{},{}".format(v, k)])
                    val.append(vessel_capacity[k, p])
                for q in products_S[p]:
                    ind.append(name2idx["L{},{},{},{}".format(i, j, v, q)])
                    val.append(-1)
                problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                          val=val)],
                                               senses=["G"],
                                               rhs=[0],
                                               names=["LOAD11_I_{}_J_{}_V_{}_P_{}".format(i, j, v, p)])

    # Constraints 12
    for i in ports:
        for j in nodes:
            for v in voyages:
                for p in products:
                    if (i, j, v) in Xvar:
                        ind = []
                        val = []
                        ind.append(name2idx["L{},{},{},{}".format(i, j, v, p)])
                        val.append(1)
                        ind.append(name2idx["X{},{},{}".format(i, j, v)])
                        val.append((-1) * max_capacity[p])
                        problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                                  val=val)],
                                                       senses=["L"],
                                                       rhs=[0],
                                                       names=["LOAD12_I_{}_J_{}_V_{}_P_{}".format(i, j, v, p)])

    # Constraints 13
    for i in nodes:
        for v in vessels:
            for p in products:
                ind = []
                val = []
                for j in nodes:
                    if (j, i, v, p) in Lvar:
                        ind.append(name2idx["L{},{},{},{}".format(j, i, v, p)])
                        val.append(1)
                    if (i, j, v, p) in Lvar:
                        ind.append(name2idx["L{},{},{},{}".format(i, j, v, p)])
                        val.append(-1)
                if i in ports:
                    for c in contracts_L[i]:
                        if (v, c, p) in Qvar:
                            ind.append(name2idx["Q{},{},{}".format(v, c, p)])
                            val.append(1)
                    for c in contracts_U[i]:
                        if (v, c, p) in Qvar:
                            ind.append(name2idx["Q{},{},{}".format(v, c, p)])
                            val.append(-1)
                problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                          val=val)],
                                               senses=["E"],
                                               rhs=[0],
                                               names=["LOAD_I_{}_V_{}_P_{}".format(i, v, p)])

    # -----------------------------------------------------------------------------------
    # Pickup Constraints
    # -----------------------------------------------------------------------------------
    # Constraints 15
    problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)
                                               for v in voyages],
                                              [1] * voyage_count] for c in evenly_spread_contracts],
                                   senses=["L"] * len(evenly_spread_contracts),
                                   rhs=[maximum_pickups[c] for c in evenly_spread_contracts])

    problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)
                                               for v in voyages],
                                              [1] * voyage_count] for c in evenly_spread_contracts],
                                   senses=["G"] * len(evenly_spread_contracts),
                                   rhs=[minimum_pickups[c] for c in evenly_spread_contracts])

    # Constraints 16
    problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)] +
                                              ["X{},{},{}".format(i, load_port[c], v) for i in nodes if
                                               (i, load_port[c]) in arc],
                                              [1] +
                                              [-1 for i in nodes if (i, load_port[c]) in arc]] for c in
                                             contracts for v in voyages],
                                   senses=["L"] * contract_count * voyage_count,
                                   rhs=[0] * contract_count * voyage_count)

    # Constraints 17
    problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)] +
                                              ["X{},{},{}".format(i, unload_port[c], v) for i in nodes if
                                               (i, unload_port[c]) in arc],
                                              [1] +
                                              [-1 for i in nodes if (i, unload_port[c]) in arc]] for c in
                                             contracts for v in voyages],
                                   senses=["L"] * contract_count * voyage_count,
                                   rhs=[0] * contract_count * voyage_count)

    # -----------------------------------------------------------------------------------
    # Loading Constraints
    # -----------------------------------------------------------------------------------
    # Constraints 18
    problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["D{},{}".format(v, c)]] +
                                                                  [name2idx["Q{},{},{}".format(v, c, p)]],
                                                              val=[(-1) * maximum_quantity[c, p]] +
                                                                  [1]) for (v, c, p) in Qvar if
                                             (c, p) in maximum_quantity],
                                   senses=["L" for (v, c, p) in Qvar if (c, p) in maximum_quantity],
                                   rhs=[0 for (v, c, p) in Qvar if (c, p) in maximum_quantity])

    problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=[name2idx["D{},{}".format(v, c)]] +
                                                                  [name2idx["Q{},{},{}".format(v, c, p)]],
                                                              val=[(-1) * minimum_quantity[c, p]] +
                                                                  [1]) for (v, c, p) in Qvar if
                                             (c, p) in minimum_quantity],
                                   senses=["G" for (v, c, p) in Qvar if (c, p) in minimum_quantity],
                                   rhs=[0 for (v, c, p) in Qvar if (c, p) in minimum_quantity])

    # Constraints 19 NB: E i xpress
    problem.linear_constraints.add(
        lin_expr=[cplex.SparsePair(ind=[name2idx["Q{},{},{}".format(v, c, p)] for v in voyages],
                                   val=[1 for v in voyages]) for c in contracts for p in products if
                  (c, p) in contract_demand and contract_demand[c, p] > 0],
        senses=["E" for c in contracts for p in products if (c, p) in contract_demand and contract_demand[c, p] > 0],
        rhs=[contract_demand[c, p] for c in contracts for p in products if
             (c, p) in contract_demand and contract_demand[c, p] > 0])

    # -----------------------------------------------------------------------------------
    # Time Constraints
    # -----------------------------------------------------------------------------------
    # Constraints 20
    for v in voyages:
        for k in vessels:
            ind = []
            val = []
            ind.append(name2idx["T{},{}".format(origin[k], v)])
            val.append(-1)
            ind.append(name2idx["Y{},{}".format(v, k)])
            val.append(vessel_available_at[k])
            problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                      val=val)],
                                           senses=["E"],
                                           rhs=[0],
                                           names=["TIME20_V_{}_K_{}".format(v, k)])

    # Constraints 21
    for (i, j) in arc:
        for v in voyages:
            ind = []
            val = []
            ind.append(name2idx["T{},{}".format(i, v)])
            val.append(1)
            ind.append(name2idx["X{},{},{}".format(i, j, v)])
            val.append(node_time[i] if i in ports else 0)
            ind.append(name2idx["T{},{}".format(j, v)])
            val.append(-1)
            for s in velocities:
                for k in vessels:
                    if (i, j, v, k, s) in Wvar:
                        ind.append(name2idx["W{},{},{},{},{}".format(i, j, v, k, s)])
                        val.append(sailing_time_2[i, j, k, s])
            if i in ports:
                for c in contracts_U[i] + contracts_L[i]:
                    for p in products:
                        if (k, c, p) in Qvar:
                            ind.append(name2idx["Q{},{},{}".format(v, c, p)])
                            val.append(handling_times[p])
            problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                      val=val)],
                                           senses=["L"],
                                           rhs=[0],
                                           names=["TIME_I_{}_J_{}_V_{}".format(i, j, v)])

    # Constraints 22
    for v in voyages:
        for c in transit_contracts:
            ind = []
            val = []
            ind.append(name2idx["T{},{}".format(load_port[c], v)])
            val.append(1)
            ind.append(name2idx["T{},{}".format(unload_port[c], v)])
            val.append(-1)
            ind.append(name2idx["D{},{}".format(v, c)])
            val.append((-1) * big_T[c])
            problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                      val=val)],
                                           senses=["G"],
                                           rhs=[(-1) * (big_T[c] + contract_transit_time[c])])

    # Constraints 23
    for j in ports:
        for k in vessels:
            for v in voyages:
                ind = []
                val = []
                ind.append(name2idx["T{},{}".format(j, v)])
                val.append(1)
                ind.append(name2idx["X{},{},{}".format(origin[k], j, v)])
                val.append(big_S[j, k])
                problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                          val=val)],
                                               senses=["L"],
                                               rhs=[(big_S[j, k] + end_of_planning_horizon)])

    # Constraints 24
    for k in vessels:
        for v in voyages:
            ind = []
            val = []
            ind.append(name2idx["WT{}".format(k)])
            val.append(1)
            ind.append(name2idx["T{},{}".format(destination[k], v)])
            val.append(-1)
            ind.append(name2idx["T{},{}".format(origin[k], v)])
            val.append(1)
            for (i, j) in arc:
                if (i, j, k) in arcs and i in ports:
                    ind.append(name2idx["X{},{},{}".format(i, j, v)])
                    val.append(node_time[i])
                for s in velocities:
                    if (i, j, k) in arcs:
                        ind.append(name2idx["W{},{},{},{},{}".format(i, j, v, k, s)])
                        val.append(sailing_time_2[i, j, k, s])
            ind.append(name2idx["Y{},{}".format(v, k)])
            val.append((-1) * big_L[k])
            problem.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind,
                                                                      val=val)],
                                           senses=["G"],
                                           rhs=[(-1) * big_L[k]])

    # -----------------------------------------------------------------------------------
    # Evenly Spread Constraints
    # -----------------------------------------------------------------------------------
    # Constraints 30
    problem.linear_constraints.add(
        lin_expr=[[["Z{},{},{}".format(v, w, c) for v in voyages for w in voyages if (v, w, c) in Zvar] +
                   ["D{},{}".format(v, c) for v in voyages],
                   [1 for w in voyages for v in voyages if (v, w, c) in Zvar] +
                   [-1] * voyage_count] for c in evenly_spread_contracts],
        senses=["G"] * len(evenly_spread_contracts),
        rhs=[-1.0] * len(evenly_spread_contracts))

    # Constraints 31
    problem.linear_constraints.add(lin_expr=[[["Z{},{},{}".format(v, w, c) for w in range((v + 1), voyage_count + 1)] +
                                              ["D{},{}".format(v, c)],
                                              [1 for w in range((v + 1), voyage_count + 1)] +
                                              [-1]] for c in evenly_spread_contracts for v in voyages],
                                   senses=["L"] * len(evenly_spread_contracts) * voyage_count,
                                   rhs=[0.0] * len(evenly_spread_contracts) * voyage_count)

    # Constraints 32
    problem.linear_constraints.add(lin_expr=[[["Z{},{},{}".format(w, v, c) for w in range(1, v)] +
                                              ["D{},{}".format(v, c)],
                                              [1 for w in range(1, v)] +
                                              [-1]] for c in evenly_spread_contracts for v in voyages],
                                   senses=["L"] * len(evenly_spread_contracts) * voyage_count,
                                   rhs=[0.0] * len(evenly_spread_contracts) * voyage_count)

    # Constraints 33
    problem.linear_constraints.add(
        lin_expr=[[["O{},{}".format(n, c) for n in range(int(minimum_pickups[c]), int(maximum_pickups[c] + 1))] +
                   ["D{},{}".format(v, c) for v in voyages],
                   [n for n in range(int(minimum_pickups[c]), int(maximum_pickups[c] + 1))] +
                   [-1] * voyage_count] for c in evenly_spread_contracts],
        senses=["E"] * len(evenly_spread_contracts),
        rhs=[0.0] * len(evenly_spread_contracts))

    # Constraints 34
    problem.linear_constraints.add(
        lin_expr=[[["O{},{}".format(n, c) for n in range(int(minimum_pickups[c]), int(maximum_pickups[c] + 1))],
                   [1.0] * int(maximum_pickups[c] - minimum_pickups[c] + 1)] for c in evenly_spread_contracts],
        senses=["E"] * len(evenly_spread_contracts),
        rhs=[1.0] * len(evenly_spread_contracts))

    # Constraints 35
    problem.linear_constraints.add(lin_expr=[[["O{},{}".format(n, c) for n in range(int(minimum_pickups[c]), int(maximum_pickups[c]+1))] +
                                              ["S{}".format(c)] +
                                              ["Z{},{},{}".format(v, w, c)] +
                                              ["T{},{}".format(load_port[c], w)] +
                                              ["T{},{}".format(load_port[c], v)],
                                              [(end_of_planning_horizon)/n for n in range(int(minimum_pickups[c]), int(maximum_pickups[c]+1))] +
                                              [-1] +
                                              [big_E[c]] +
                                              [-1] +
                                              [1]]
                                             for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                                   senses=["L" for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                                   rhs=[big_E[c] for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)])

    # Constraints 36
    problem.linear_constraints.add(lin_expr=[[["O{},{}".format(n, c) for n in range(int(minimum_pickups[c]), int(maximum_pickups[c]+1))] +
                                              ["S{}".format(c)] +
                                              ["Z{},{},{}".format(v, w, c)] +
                                              ["T{},{}".format(load_port[c], w)] +
                                              ["T{},{}".format(load_port[c], v)],
                                              [(end_of_planning_horizon)/n for n in range(int(minimum_pickups[c]), int(maximum_pickups[c]+1))] +
                                              [1] +
                                              [(-1)*big_E[c]] +
                                              [-1] +
                                              [1]]
                                             for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                                   senses=["G" for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)],
                                   rhs=[(-1)*big_E[c] for c in evenly_spread_contracts for v in voyages for w in range(v+1, voyage_count+1)])

    # Constraints 37
    problem.linear_constraints.add(lin_expr=[[["S{}".format(c) for c in evenly_spread_contracts],
                                              [1.0]*(len(evenly_spread_contracts))]],
                                   senses=["G"],
                                   rhs=[service_level])

    # -----------------------------------------------------------------------------------
    # Valid Ineq
    # -----------------------------------------------------------------------------------
    # Constraints 25
    problem.linear_constraints.add(lin_expr=[[["WT{}".format(k) for k in vessels],
                                              [1]*vessel_count]],
                                   senses=["G"],
                                   rhs=[sum([2*handling_times[p]*contract_demand[c, p] for p in products for c in contracts if (c, p) in contract_demand])])

    '''problem.linear_constraints.add(lin_expr=[[["L{},{},{},{}".format(origin[k], j, v, p) for k in vessels for j in nodes if (origin[k], j, v, p) in Lvar],
                                              [1 for k in vessels for j in nodes if (origin[k], j, v, p) in Lvar]] for v in voyages for p in products],
                                   senses=["E"]*voyage_count*product_count,
                                   rhs=[0]*voyage_count*product_count)'''



    # Writing the problem to text file
    # problem.write("voyage_model.lp")
    return problem


def add_fixed_variables(problem, X, DELTA, phi, Z, y):
    problem.linear_constraints.add(lin_expr=[[["D{},{}".format(v, c)],
                                                  [1]] for (v, c) in Dvar],
                                   senses=["E" for d in Dvar],
                                   rhs=[(DELTA[v, c] if (v, c) in DELTA and DELTA[v, c] == 1 else 0) for (v, c) in Dvar])

    problem.linear_constraints.add(lin_expr=[[["Z{},{},{}".format(v, w, c)],
                                                  [1]] for (v, w, c) in Zvar],
                                   senses=["E" for (v, w, c) in Zvar],
                                   rhs=[Z[v, w, c] if (v, w, c) in Z else 0 for (v, w, c) in Zvar])

    problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(key[0], key[1], key[2])],
                                              [1]] for key in Xvar],
                                   senses=["E" for key in Xvar],
                                   rhs=[X[key] if key in X else 0 for key in Xvar])

    '''problem.linear_constraints.add(lin_expr=[[["X{},{},{}".format(key[0], key[1], key[2])],
                                              [1]] for key, value in X.items()],
                                   senses=["E" *len(X)],
                                   rhs=[value for key, value in X.items()])'''



    problem.linear_constraints.add(lin_expr=[[["O{},{}".format(n, c)],
                                                  [1]] for (n, c) in Ovar],
                                   senses=["E" for (n, c) in Ovar],
                                   rhs=[phi[n, c] if (n, c) in phi else 0 for (n, c) in Ovar])

    problem.linear_constraints.add(lin_expr=[[["Y{},{}".format(key[0], key[1])],
                                                  [1]] for key in Yvar],
                                   senses=["E" for key in Yvar],
                                   rhs=[y[key] if key in y else 0 for key in Yvar])

    '''problem.set_log_stream(None)
    problem.set_error_stream(None)
    problem.set_warning_stream(None)
    problem.set_results_stream(None)'''

    return problem